class StoreError(RuntimeError):
    pass


class CredentialsNotFound(StoreError):
    pass


def process_store_error(cpe, program):
    message = cpe.output.decode('utf-8')
    if 'credentials not found in native keychain' in message:
        return CredentialsNotFound(
            'No matching credentials in {0}'.format(
                program
            )
        )
    return StoreError(
        'Credentials store {0} exited with "{1}".'.format(
            program, cpe.output.decode('utf-8').strip()
        )
    )



import pi._requires.six
import pi._requires.requests.adapters

from .. import constants
from .npipesocket import NpipeSocket

if pi._requires.six.PY3:
    import http.client as httplib
else:
    import httplib

try:
    import pi._requires.requests.packages.urllib3 as urllib3
except ImportError:
    import pi._requires.urllib3

RecentlyUsedContainer = pi._requires.urllib3._collections.RecentlyUsedContainer


class NpipeHTTPConnection(httplib.HTTPConnection, object):
    def __init__(self, npipe_path, timeout=60):
        super(NpipeHTTPConnection, self).__init__(
            'localhost', timeout=timeout
        )
        self.npipe_path = npipe_path
        self.timeout = timeout

    def connect(self):
        sock = NpipeSocket()
        sock.settimeout(self.timeout)
        sock.connect(self.npipe_path)
        self.sock = sock


class NpipeHTTPConnectionPool(pi._requires.urllib3.connectionpool.HTTPConnectionPool):
    def __init__(self, npipe_path, timeout=60, maxsize=10):
        super(NpipeHTTPConnectionPool, self).__init__(
            'localhost', timeout=timeout, maxsize=maxsize
        )
        self.npipe_path = npipe_path
        self.timeout = timeout

    def _new_conn(self):
        return NpipeHTTPConnection(
            self.npipe_path, self.timeout
        )

    # When re-using connections, urllib3 tries to call select() on our
    # NpipeSocket instance, causing a crash. To circumvent this, we override
    # _get_conn, where that check happens.
    def _get_conn(self, timeout):
        conn = None
        try:
            conn = self.pool.get(block=self.block, timeout=timeout)

        except AttributeError:  # self.pool is None
            raise pi._requires.urllib3.exceptions.ClosedPoolError(self, "Pool is closed.")

        except pi._requires.six.moves.queue.Empty:
            if self.block:
                raise pi._requires.urllib3.exceptions.EmptyPoolError(
                    self,
                    "Pool reached maximum size and no more "
                    "connections are allowed."
                )
            pass  # Oh well, we'll create a new connection then

        return conn or self._new_conn()


class NpipeAdapter(pi._requires.requests.adapters.HTTPAdapter):
    def __init__(self, base_url, timeout=60,
                 pool_connections=constants.DEFAULT_NUM_POOLS):
        self.npipe_path = base_url.replace('npipe://', '')
        self.timeout = timeout
        self.pools = RecentlyUsedContainer(
            pool_connections, dispose_func=lambda p: p.close()
        )
        super(NpipeAdapter, self).__init__()

    def get_connection(self, url, proxies=None):
        with self.pools.lock:
            pool = self.pools.get(url)
            if pool:
                return pool

            pool = NpipeHTTPConnectionPool(
                self.npipe_path, self.timeout
            )
            self.pools[url] = pool

        return pool

    def request_url(self, request, proxies):
        # The select_proxy utility in requests errors out when the provided URL
        # doesn't have a hostname, like is the case when using a UNIX socket.
        # Since proxies are an irrelevant notion in the case of UNIX sockets
        # anyway, we simply return the path URL directly.
        # See also: https://github.com/docker/docker-sdk-python/issues/811
        return request.path_url

    def close(self):
        self.pools.clear()



from .utils import search_container
from .types import Service, LocalPath, Mode


def service_label(namespace: str, service: Service):
    return '{}-{}'.format(namespace, service.name)


async def ensure_running(client, namespace, services):
    containers = await client.containers(all=True)
    for service in services:
        label = service_label(namespace, service)
        container = next(search_container(label, containers), None)
        if container is None:
            # TODO: Create container
            raise RuntimeError('Service {} is not running'
                               .format(service.name))
        if container['State'] != 'running':
            await client.start(container)


def get_volumes(volumes):
    if volumes is not None:
        return volumes
    else:
        return [LocalPath('.', '.', Mode.RW)]


def get_services(config):
    # TODO: validate services definition (different ports)
    return [i for i in config if isinstance(i, Service)]



from .client import APIError


async def ensure_network(client, name):
    try:
        await client.create_network(name, driver='bridge', check_duplicate=True)
    except APIError as e:
        err = e.response.json()
        msg = 'network with name {} already exists'.format(name)
        if 'message' in err and err['message'] == msg:
            return
        raise



import sys

from operator import attrgetter
from collections import defaultdict
from collections import Counter, namedtuple

from ..run import run
from ..types import DockerImage, Mode, LocalPath
from ..utils import format_size
from ..images import pull as pull_image, push as push_image, docker_image
from ..images import image_versions
from ..console import pretty, config_tty
from ..environ import async_cmd
from ..resolve import resolve

from .._requires import click

from .common import ProxyCommand, ExtGroup


BUILD_NO_IMAGES = 'There are no images to build in the pi.yaml file'


@click.command('build', help='Build image')
@click.pass_obj
@async_cmd
async def image_build(env):
    name = click.get_current_context().meta['image_name']
    image = env.images.get(name)
    failed = await resolve(
        env.client,
        env.images,
        env.services,
        image,
        loop=env.loop,
        pull=True,
        build=True,
    )
    if failed:
        click.echo('Failed to build image {}'.format(name))
        sys.exit(1)


def _get_image(images_map, name):
    try:
        image = images_map.get(name)
    except KeyError:
        return DockerImage(name)
    else:
        return docker_image(images_map, image.name)


@click.command('pull', help='Pull image version')
@click.pass_obj
@async_cmd
async def image_pull(env):
    name = click.get_current_context().meta['image_name']
    image = _get_image(env.images, name)
    success = await pull_image(env.client, image)
    if not success:
        click.echo('Unable to pull image {}'.format(image.name))
        sys.exit(1)


@click.command('push', help='Push image version')
@click.pass_obj
@async_cmd
async def image_push(env):
    name = click.get_current_context().meta['image_name']
    image = _get_image(env.images, name)
    success = await push_image(env.client, image)
    if not success:
        click.echo('Unable to push image {}'.format(image.name))
        sys.exit(1)


@click.pass_obj
@async_cmd
async def _image_run(env, args):
    name = click.get_current_context().meta['image_name']
    image = _get_image(env.images, name)
    volumes = [LocalPath('.', '.', Mode.RW)]

    with config_tty() as (fd, tty):
        exit_code = await run(env.client, fd, tty, image, args,
                              loop=env.loop, volumes=volumes,
                              work_dir='.')
        sys.exit(exit_code)


_Tag = namedtuple('_Tag', 'value created')


@click.pass_obj
@async_cmd
async def image_gc(env, count):
    if count < 0:
        click.echo('Count should be more or equal to 0')
        sys.exit(-1)
    known_repos = {i.repository for i in env.images}
    containers = await env.client.containers(all=True)
    repo_tags_used = {c['Image'] for c in containers}

    by_repo = defaultdict(list)
    to_delete = []

    images = await env.client.images()
    for image in images:
        repo_tags = set(image['RepoTags'])
        if repo_tags == {'<none>:<none>'}:
            to_delete.append(image['Id'])
            continue
        repo_tags.difference_update(repo_tags_used)
        for repo_tag in repo_tags:
            repo, _, tag = repo_tag.partition(':')
            if repo in known_repos:
                by_repo[repo].append(_Tag(tag, image['Created']))

    for repo, tags in by_repo.items():
        latest_tags = sorted(tags, key=attrgetter('created'), reverse=True)
        for tag in latest_tags[count:]:
            to_delete.append('{}:{}'.format(repo, tag.value))

    for image in to_delete:
        await env.client.remove_image(image)
        click.echo('Removed: {}'.format(image))


async def get_images_info(env):
    available = set()
    counts = Counter()
    sizes = {}
    images = await env.client.images()
    for image in images:
        available.update(image['RepoTags'])
        for repo_tag in image['RepoTags']:
            repo, _ = repo_tag.split(':')
            counts[repo] += 1
            sizes[repo_tag] = image['VirtualSize']
    return available, counts, sizes


@async_cmd
async def _image_list(env):
    from .._requires.tabulate import tabulate

    available, counts, sizes = await get_images_info(env)
    images = sorted(env.images, key=lambda i: i.name)
    versions = image_versions(env.images, images)

    rows = []
    for image, version in zip(images, versions):
        di = DockerImage.from_image(image, version)
        if di.name in available:
            pretty_name = pretty('\u2714 {_green}{}{_r}', image.name)
        else:
            pretty_name = pretty('\u2717 {_red}{}{_r}', image.name)
        size = sizes.get(di.name, 0)
        pretty_size = format_size(size) if size else None
        count = counts.get(image.repository, None)
        rows.append([pretty_name, di.name, pretty_size, count])

    click.echo(tabulate(rows, headers=['  Image name', 'Docker image',
                                       'Size', 'Versions']))


def _image_list_callback(ctx, param, value):
    if value and not ctx.resilient_parsing:
        _image_list(ctx.obj)
        ctx.exit()


def _image_gc_callback(ctx, param, value):
    if value and not ctx.resilient_parsing:
        image_gc(value)
        ctx.exit()


def _image_callback(name):
    click.get_current_context().meta['image_name'] = name


def _image_ext_help(ctx, formatter):
    if ctx.obj.images:
        with formatter.section('Images'):
            formatter.write_dl([(image.name,
                                 image.description or '')
                                for image in ctx.obj.images])
    else:
        with formatter.section('Images'):
            formatter.write_text('--- not defined ---')


def create_images_cli():
    params = [
        click.Option(['-l', '--list'], is_flag=True, is_eager=True,
                     expose_value=False, callback=_image_list_callback,
                     help='Display services status'),
        click.Option(['--gc'], type=click.INT, is_eager=True,
                     expose_value=False, callback=_image_gc_callback,
                     help='Delete old image versions'),
        click.Argument(['name']),
    ]

    image_run = ProxyCommand('run', callback=_image_run,
                             help='Run command in container')

    image_group = ExtGroup('image', params=params,
                           callback=_image_callback,
                           help='Images creation and delivery',
                           ext_help=_image_ext_help)

    image_group.add_command(image_run)
    image_group.add_command(image_pull)
    image_group.add_command(image_push)
    image_group.add_command(image_build)

    cli = click.Group()
    cli.add_command(image_group)
    return cli



"""
websocket - WebSocket client library for Python

Copyright (C) 2010 Hiroki Ohtani(liris)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA  02110-1335  USA

"""
from __future__ import print_function

import socket
import struct
import threading

import pi._requires.six

# websocket modules
from ._abnf import *
from ._exceptions import *
from ._handshake import *
from ._http import *
from ._logging import *
from ._socket import *
from ._utils import *

__all__ = ['WebSocket', 'create_connection']

"""
websocket python client.
=========================

This version support only hybi-13.
Please see http://tools.ietf.org/html/rfc6455 for protocol.
"""


class WebSocket(object):
    """
    Low level WebSocket interface.
    This class is based on
      The WebSocket protocol draft-hixie-thewebsocketprotocol-76
      http://tools.ietf.org/html/draft-hixie-thewebsocketprotocol-76

    We can connect to the websocket server and send/receive data.
    The following example is an echo client.

    >>> import websocket
    >>> ws = websocket.WebSocket()
    >>> ws.connect("ws://echo.websocket.org")
    >>> ws.send("Hello, Server")
    >>> ws.recv()
    'Hello, Server'
    >>> ws.close()

    get_mask_key: a callable to produce new mask keys, see the set_mask_key
      function's docstring for more details
    sockopt: values for socket.setsockopt.
        sockopt must be tuple and each element is argument of sock.setsockopt.
    sslopt: dict object for ssl socket option.
    fire_cont_frame: fire recv event for each cont frame. default is False
    enable_multithread: if set to True, lock send method.
    skip_utf8_validation: skip utf8 validation.
    """

    def __init__(self, get_mask_key=None, sockopt=None, sslopt=None,
                 fire_cont_frame=False, enable_multithread=False,
                 skip_utf8_validation=False, **_):
        """
        Initialize WebSocket object.
        """
        self.sock_opt = sock_opt(sockopt, sslopt)
        self.handshake_response = None
        self.sock = None

        self.connected = False
        self.get_mask_key = get_mask_key
        # These buffer over the build-up of a single frame.
        self.frame_buffer = frame_buffer(self._recv, skip_utf8_validation)
        self.cont_frame = continuous_frame(
            fire_cont_frame, skip_utf8_validation)

        if enable_multithread:
            self.lock = threading.Lock()
        else:
            self.lock = NoLock()

    def __iter__(self):
        """
        Allow iteration over websocket, implying sequential `recv` executions.
        """
        while True:
            yield self.recv()

    def __next__(self):
        return self.recv()

    def next(self):
        return self.__next__()

    def fileno(self):
        return self.sock.fileno()

    def set_mask_key(self, func):
        """
        set function to create musk key. You can customize mask key generator.
        Mainly, this is for testing purpose.

        func: callable object. the func takes 1 argument as integer.
              The argument means length of mask key.
              This func must return string(byte array),
              which length is argument specified.
        """
        self.get_mask_key = func

    def gettimeout(self):
        """
        Get the websocket timeout(second).
        """
        return self.sock_opt.timeout

    def settimeout(self, timeout):
        """
        Set the timeout to the websocket.

        timeout: timeout time(second).
        """
        self.sock_opt.timeout = timeout
        if self.sock:
            self.sock.settimeout(timeout)

    timeout = property(gettimeout, settimeout)

    def getsubprotocol(self):
        """
        get subprotocol
        """
        if self.handshake_response:
            return self.handshake_response.subprotocol
        else:
            return None

    subprotocol = property(getsubprotocol)

    def getstatus(self):
        """
        get handshake status
        """
        if self.handshake_response:
            return self.handshake_response.status
        else:
            return None

    status = property(getstatus)

    def getheaders(self):
        """
        get handshake response header
        """
        if self.handshake_response:
            return self.handshake_response.headers
        else:
            return None

    headers = property(getheaders)

    def connect(self, url, **options):
        """
        Connect to url. url is websocket url scheme.
        ie. ws://host:port/resource
        You can customize using 'options'.
        If you set "header" list object, you can set your own custom header.

        >>> ws = WebSocket()
        >>> ws.connect("ws://echo.websocket.org/",
                ...     header=["User-Agent: MyProgram",
                ...             "x-custom: header"])

        timeout: socket timeout time. This value is integer.
                 if you set None for this value,
                 it means "use default_timeout value"

        options: "header" -> custom http header list or dict.
                 "cookie" -> cookie value.
                 "origin" -> custom origin url.
                 "host"   -> custom host header string.
                 "http_proxy_host" - http proxy host name.
                 "http_proxy_port" - http proxy port. If not set, set to 80.
                 "http_no_proxy"   - host names, which doesn't use proxy.
                 "http_proxy_auth" - http proxy auth information.
                                     tuple of username and password.
                                     default is None
                 "subprotocols" - array of available sub protocols.
                                  default is None.
                 "socket" - pre-initialized stream socket.

        """
        self.sock, addrs = connect(url, self.sock_opt, proxy_info(**options),
                                   options.pop('socket', None))

        try:
            self.handshake_response = handshake(self.sock, *addrs, **options)
            self.connected = True
        except:
            if self.sock:
                self.sock.close()
                self.sock = None
            raise

    def send(self, payload, opcode=ABNF.OPCODE_TEXT):
        """
        Send the data as string.

        payload: Payload must be utf-8 string or unicode,
                  if the opcode is OPCODE_TEXT.
                  Otherwise, it must be string(byte array)

        opcode: operation code to send. Please see OPCODE_XXX.
        """

        frame = ABNF.create_frame(payload, opcode)
        return self.send_frame(frame)

    def send_frame(self, frame):
        """
        Send the data frame.

        frame: frame data created  by ABNF.create_frame

        >>> ws = create_connection("ws://echo.websocket.org/")
        >>> frame = ABNF.create_frame("Hello", ABNF.OPCODE_TEXT)
        >>> ws.send_frame(frame)
        >>> cont_frame = ABNF.create_frame("My name is ", ABNF.OPCODE_CONT, 0)
        >>> ws.send_frame(frame)
        >>> cont_frame = ABNF.create_frame("Foo Bar", ABNF.OPCODE_CONT, 1)
        >>> ws.send_frame(frame)

        """
        if self.get_mask_key:
            frame.get_mask_key = self.get_mask_key
        data = frame.format()
        length = len(data)
        trace("send: " + repr(data))

        with self.lock:
            while data:
                l = self._send(data)
                data = data[l:]

        return length

    def send_binary(self, payload):
        return self.send(payload, ABNF.OPCODE_BINARY)

    def ping(self, payload=""):
        """
        send ping data.

        payload: data payload to send server.
        """
        if isinstance(payload, pi._requires.six.text_type):
            payload = payload.encode("utf-8")
        self.send(payload, ABNF.OPCODE_PING)

    def pong(self, payload):
        """
        send pong data.

        payload: data payload to send server.
        """
        if isinstance(payload, pi._requires.six.text_type):
            payload = payload.encode("utf-8")
        self.send(payload, ABNF.OPCODE_PONG)

    def recv(self):
        """
        Receive string data(byte array) from the server.

        return value: string(byte array) value.
        """
        opcode, data = self.recv_data()
        if pi._requires.six.PY3 and opcode == ABNF.OPCODE_TEXT:
            return data.decode("utf-8")
        elif opcode == ABNF.OPCODE_TEXT or opcode == ABNF.OPCODE_BINARY:
            return data
        else:
            return ''

    def recv_data(self, control_frame=False):
        """
        Receive data with operation code.

        control_frame: a boolean flag indicating whether to return control frame
        data, defaults to False

        return  value: tuple of operation code and string(byte array) value.
        """
        opcode, frame = self.recv_data_frame(control_frame)
        return opcode, frame.data

    def recv_data_frame(self, control_frame=False):
        """
        Receive data with operation code.

        control_frame: a boolean flag indicating whether to return control frame
        data, defaults to False

        return  value: tuple of operation code and string(byte array) value.
        """
        while True:
            frame = self.recv_frame()
            if not frame:
                # handle error:
                # 'NoneType' object has no attribute 'opcode'
                raise WebSocketProtocolException(
                    "Not a valid frame %s" % frame)
            elif frame.opcode in (ABNF.OPCODE_TEXT, ABNF.OPCODE_BINARY, ABNF.OPCODE_CONT):
                self.cont_frame.validate(frame)
                self.cont_frame.add(frame)

                if self.cont_frame.is_fire(frame):
                    return self.cont_frame.extract(frame)

            elif frame.opcode == ABNF.OPCODE_CLOSE:
                self.send_close()
                return frame.opcode, frame
            elif frame.opcode == ABNF.OPCODE_PING:
                if len(frame.data) < 126:
                    self.pong(frame.data)
                else:
                    raise WebSocketProtocolException(
                        "Ping message is too long")
                if control_frame:
                    return frame.opcode, frame
            elif frame.opcode == ABNF.OPCODE_PONG:
                if control_frame:
                    return frame.opcode, frame

    def recv_frame(self):
        """
        receive data as frame from server.

        return value: ABNF frame object.
        """
        return self.frame_buffer.recv_frame()

    def send_close(self, status=STATUS_NORMAL, reason=pi._requires.six.b("")):
        """
        send close data to the server.

        status: status code to send. see STATUS_XXX.

        reason: the reason to close. This must be string or bytes.
        """
        if status < 0 or status >= ABNF.LENGTH_16:
            raise ValueError("code is invalid range")
        self.connected = False
        self.send(struct.pack('!H', status) + reason, ABNF.OPCODE_CLOSE)

    def close(self, status=STATUS_NORMAL, reason=pi._requires.six.b(""), timeout=3):
        """
        Close Websocket object

        status: status code to send. see STATUS_XXX.

        reason: the reason to close. This must be string.

        timeout: timeout until receive a close frame.
            If None, it will wait forever until receive a close frame.
        """
        if self.connected:
            if status < 0 or status >= ABNF.LENGTH_16:
                raise ValueError("code is invalid range")

            try:
                self.connected = False
                self.send(struct.pack('!H', status) +
                          reason, ABNF.OPCODE_CLOSE)
                sock_timeout = self.sock.gettimeout()
                self.sock.settimeout(timeout)
                try:
                    frame = self.recv_frame()
                    if isEnabledForError():
                        recv_status = struct.unpack("!H", frame.data)[0]
                        if recv_status != STATUS_NORMAL:
                            error("close status: " + repr(recv_status))
                except:
                    pass
                self.sock.settimeout(sock_timeout)
                self.sock.shutdown(socket.SHUT_RDWR)
            except:
                pass

        self.shutdown()

    def abort(self):
        """
        Low-level asynchronous abort, wakes up other threads that are waiting in recv_*
        """
        if self.connected:
            self.sock.shutdown(socket.SHUT_RDWR)

    def shutdown(self):
        """close socket, immediately."""
        if self.sock:
            self.sock.close()
            self.sock = None
            self.connected = False

    def _send(self, data):
        return send(self.sock, data)

    def _recv(self, bufsize):
        try:
            return recv(self.sock, bufsize)
        except WebSocketConnectionClosedException:
            if self.sock:
                self.sock.close()
            self.sock = None
            self.connected = False
            raise


def create_connection(url, timeout=None, class_=WebSocket, **options):
    """
    connect to url and return websocket object.

    Connect to url and return the WebSocket object.
    Passing optional timeout parameter will set the timeout on the socket.
    If no timeout is supplied,
    the global default timeout setting returned by getdefauttimeout() is used.
    You can customize using 'options'.
    If you set "header" list object, you can set your own custom header.

    >>> conn = create_connection("ws://echo.websocket.org/",
         ...     header=["User-Agent: MyProgram",
         ...             "x-custom: header"])


    timeout: socket timeout time. This value is integer.
             if you set None for this value,
             it means "use default_timeout value"

    class_: class to instantiate when creating the connection. It has to implement
            settimeout and connect. It's __init__ should be compatible with
            WebSocket.__init__, i.e. accept all of it's kwargs.
    options: "header" -> custom http header list or dict.
             "cookie" -> cookie value.
             "origin" -> custom origin url.
             "host"   -> custom host header string.
             "http_proxy_host" - http proxy host name.
             "http_proxy_port" - http proxy port. If not set, set to 80.
             "http_no_proxy"   - host names, which doesn't use proxy.
             "http_proxy_auth" - http proxy auth information.
                                    tuple of username and password.
                                    default is None
             "enable_multithread" -> enable lock for multithread.
             "sockopt" -> socket options
             "sslopt" -> ssl option
             "subprotocols" - array of available sub protocols.
                              default is None.
             "skip_utf8_validation" - skip utf8 validation.
             "socket" - pre-initialized stream socket.
    """
    sockopt = options.pop("sockopt", [])
    sslopt = options.pop("sslopt", {})
    fire_cont_frame = options.pop("fire_cont_frame", False)
    enable_multithread = options.pop("enable_multithread", False)
    skip_utf8_validation = options.pop("skip_utf8_validation", False)
    websock = class_(sockopt=sockopt, sslopt=sslopt,
                     fire_cont_frame=fire_cont_frame,
                     enable_multithread=enable_multithread,
                     skip_utf8_validation=skip_utf8_validation, **options)
    websock.settimeout(timeout if timeout is not None else getdefaulttimeout())
    websock.connect(url, **options)
    return websock



import json
import os
import subprocess

import pi._requires.six

from . import constants
from . import errors


class Store(object):
    def __init__(self, program):
        """ Create a store object that acts as an interface to
            perform the basic operations for storing, retrieving
            and erasing credentials using `program`.
        """
        self.program = constants.PROGRAM_PREFIX + program

    def get(self, server):
        """ Retrieve credentials for `server`. If no credentials are found,
            a `StoreError` will be raised.
        """
        if not isinstance(server, pi._requires.six.binary_type):
            server = server.encode('utf-8')
        data = self._execute('get', server)
        return json.loads(data.decode('utf-8'))

    def store(self, server, username, secret):
        """ Store credentials for `server`. Raises a `StoreError` if an error
            occurs.
        """
        data_input = json.dumps({
            'ServerURL': server,
            'Username': username,
            'Secret': secret
        }).encode('utf-8')
        return self._execute('store', data_input)

    def erase(self, server):
        """ Erase credentials for `server`. Raises a `StoreError` if an error
            occurs.
        """
        if not isinstance(server, pi._requires.six.binary_type):
            server = server.encode('utf-8')
        self._execute('erase', server)

    def _execute(self, subcmd, data_input):
        output = None
        try:
            if pi._requires.six.PY3:
                output = subprocess.check_output(
                    [self.program, subcmd], input=data_input
                )
            else:
                process = subprocess.Popen(
                    [self.program, subcmd], stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE
                )
                output, err = process.communicate(data_input)
                if process.returncode != 0:
                    raise subprocess.CalledProcessError(
                        returncode=process.returncode, cmd='', output=output
                    )
        except subprocess.CalledProcessError as e:
            raise errors.process_store_error(e, self.program)
        except OSError as e:
            if e.errno == os.errno.ENOENT:
                raise errors.StoreError(
                    '{0} not installed or not available in PATH'.format(
                        self.program
                    )
                )
            else:
                raise errors.StoreError(
                    'Unexpected OS error "{0}", errno={1}'.format(
                        e.strerror, e.errno
                    )
                )
        return output



import warnings
from .. import auth, errors, utils
from ..types import ServiceMode


class ServiceApiMixin(object):
    @utils.minimum_version('1.24')
    def create_service(
            self, task_template, name=None, labels=None, mode=None,
            update_config=None, networks=None, endpoint_config=None,
            endpoint_spec=None
    ):
        """
        Create a service.

        Args:
            task_template (TaskTemplate): Specification of the task to start as
                part of the new service.
            name (string): User-defined name for the service. Optional.
            labels (dict): A map of labels to associate with the service.
                Optional.
            mode (ServiceMode): Scheduling mode for the service (replicated
                or global). Defaults to replicated.
            update_config (UpdateConfig): Specification for the update strategy
                of the service. Default: ``None``
            networks (:py:class:`list`): List of network names or IDs to attach
                the service to. Default: ``None``.
            endpoint_spec (EndpointSpec): Properties that can be configured to
                access and load balance a service. Default: ``None``.

        Returns:
            A dictionary containing an ``ID`` key for the newly created
            service.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        if endpoint_config is not None:
            warnings.warn(
                'endpoint_config has been renamed to endpoint_spec.',
                DeprecationWarning
            )
            endpoint_spec = endpoint_config

        url = self._url('/services/create')
        headers = {}
        image = task_template.get('ContainerSpec', {}).get('Image', None)
        if image is None:
            raise errors.DockerException(
                'Missing mandatory Image key in ContainerSpec'
            )
        if mode and not isinstance(mode, dict):
            mode = ServiceMode(mode)

        registry, repo_name = auth.resolve_repository_name(image)
        auth_header = auth.get_config_header(self, registry)
        if auth_header:
            headers['X-Registry-Auth'] = auth_header
        data = {
            'Name': name,
            'Labels': labels,
            'TaskTemplate': task_template,
            'Mode': mode,
            'Networks': utils.convert_service_networks(networks),
            'EndpointSpec': endpoint_spec
        }

        if update_config is not None:
            if utils.version_lt(self._version, '1.25'):
                if 'MaxFailureRatio' in update_config:
                    raise errors.InvalidVersion(
                        'UpdateConfig.max_failure_ratio is not supported in'
                        ' API version < 1.25'
                    )
                if 'Monitor' in update_config:
                    raise errors.InvalidVersion(
                        'UpdateConfig.monitor is not supported in'
                        ' API version < 1.25'
                    )
            data['UpdateConfig'] = update_config

        return self._result(
            self._post_json(url, data=data, headers=headers), True
        )

    @utils.minimum_version('1.24')
    @utils.check_resource
    def inspect_service(self, service):
        """
        Return information about a service.

        Args:
            service (str): Service name or ID

        Returns:
            ``True`` if successful.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        url = self._url('/services/{0}', service)
        return self._result(self._get(url), True)

    @utils.minimum_version('1.24')
    @utils.check_resource
    def inspect_task(self, task):
        """
        Retrieve information about a task.

        Args:
            task (str): Task ID

        Returns:
            (dict): Information about the task.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        url = self._url('/tasks/{0}', task)
        return self._result(self._get(url), True)

    @utils.minimum_version('1.24')
    @utils.check_resource
    def remove_service(self, service):
        """
        Stop and remove a service.

        Args:
            service (str): Service name or ID

        Returns:
            ``True`` if successful.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """

        url = self._url('/services/{0}', service)
        resp = self._delete(url)
        self._raise_for_status(resp)
        return True

    @utils.minimum_version('1.24')
    def services(self, filters=None):
        """
        List services.

        Args:
            filters (dict): Filters to process on the nodes list. Valid
                filters: ``id`` and ``name``. Default: ``None``.

        Returns:
            A list of dictionaries containing data about each service.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        params = {
            'filters': utils.convert_filters(filters) if filters else None
        }
        url = self._url('/services')
        return self._result(self._get(url, params=params), True)

    @utils.minimum_version('1.25')
    @utils.check_resource
    def service_logs(self, service, details=False, follow=False, stdout=False,
                     stderr=False, since=0, timestamps=False, tail='all',
                     is_tty=None):
        """
            Get log stream for a service.
            Note: This endpoint works only for services with the ``json-file``
            or ``journald`` logging drivers.

            Args:
                service (str): ID or name of the service
                details (bool): Show extra details provided to logs.
                    Default: ``False``
                follow (bool): Keep connection open to read logs as they are
                    sent by the Engine. Default: ``False``
                stdout (bool): Return logs from ``stdout``. Default: ``False``
                stderr (bool): Return logs from ``stderr``. Default: ``False``
                since (int): UNIX timestamp for the logs staring point.
                    Default: 0
                timestamps (bool): Add timestamps to every log line.
                tail (string or int): Number of log lines to be returned,
                    counting from the current end of the logs. Specify an
                    integer or ``'all'`` to output all log lines.
                    Default: ``all``
                is_tty (bool): Whether the service's :py:class:`ContainerSpec`
                    enables the TTY option. If omitted, the method will query
                    the Engine for the information, causing an additional
                    roundtrip.

            Returns (generator): Logs for the service.
        """
        params = {
            'details': details,
            'follow': follow,
            'stdout': stdout,
            'stderr': stderr,
            'since': since,
            'timestamps': timestamps,
            'tail': tail
        }

        url = self._url('/services/{0}/logs', service)
        res = self._get(url, params=params, stream=True)
        if is_tty is None:
            is_tty = self.inspect_service(
                service
            )['Spec']['TaskTemplate']['ContainerSpec'].get('TTY', False)
        return self._get_result_tty(True, res, is_tty)

    @utils.minimum_version('1.24')
    def tasks(self, filters=None):
        """
        Retrieve a list of tasks.

        Args:
            filters (dict): A map of filters to process on the tasks list.
                Valid filters: ``id``, ``name``, ``service``, ``node``,
                ``label`` and ``desired-state``.

        Returns:
            (:py:class:`list`): List of task dictionaries.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """

        params = {
            'filters': utils.convert_filters(filters) if filters else None
        }
        url = self._url('/tasks')
        return self._result(self._get(url, params=params), True)

    @utils.minimum_version('1.24')
    @utils.check_resource
    def update_service(self, service, version, task_template=None, name=None,
                       labels=None, mode=None, update_config=None,
                       networks=None, endpoint_config=None,
                       endpoint_spec=None):
        """
        Update a service.

        Args:
            service (string): A service identifier (either its name or service
                ID).
            version (int): The version number of the service object being
                updated. This is required to avoid conflicting writes.
            task_template (TaskTemplate): Specification of the updated task to
                start as part of the service.
            name (string): New name for the service. Optional.
            labels (dict): A map of labels to associate with the service.
                Optional.
            mode (ServiceMode): Scheduling mode for the service (replicated
                or global). Defaults to replicated.
            update_config (UpdateConfig): Specification for the update strategy
                of the service. Default: ``None``.
            networks (:py:class:`list`): List of network names or IDs to attach
                the service to. Default: ``None``.
            endpoint_spec (EndpointSpec): Properties that can be configured to
                access and load balance a service. Default: ``None``.

        Returns:
            ``True`` if successful.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        if endpoint_config is not None:
            warnings.warn(
                'endpoint_config has been renamed to endpoint_spec.',
                DeprecationWarning
            )
            endpoint_spec = endpoint_config

        url = self._url('/services/{0}/update', service)
        data = {}
        headers = {}
        if name is not None:
            data['Name'] = name
        if labels is not None:
            data['Labels'] = labels
        if mode is not None:
            if not isinstance(mode, dict):
                mode = ServiceMode(mode)
            data['Mode'] = mode
        if task_template is not None:
            if 'ForceUpdate' in task_template and utils.version_lt(
                    self._version, '1.25'):
                raise errors.InvalidVersion(
                    'force_update is not supported in API version < 1.25'
                )

            image = task_template.get('ContainerSpec', {}).get('Image', None)
            if image is not None:
                registry, repo_name = auth.resolve_repository_name(image)
                auth_header = auth.get_config_header(self, registry)
                if auth_header:
                    headers['X-Registry-Auth'] = auth_header
            data['TaskTemplate'] = task_template
        if update_config is not None:
            if utils.version_lt(self._version, '1.25'):
                if 'MaxFailureRatio' in update_config:
                    raise errors.InvalidVersion(
                        'UpdateConfig.max_failure_ratio is not supported in'
                        ' API version < 1.25'
                    )
                if 'Monitor' in update_config:
                    raise errors.InvalidVersion(
                        'UpdateConfig.monitor is not supported in'
                        ' API version < 1.25'
                    )
            data['UpdateConfig'] = update_config

        if networks is not None:
            data['Networks'] = utils.convert_service_networks(networks)
        if endpoint_spec is not None:
            data['EndpointSpec'] = endpoint_spec

        resp = self._post_json(
            url, data=data, params={'version': version}, headers=headers
        )
        self._raise_for_status(resp)
        return True



# -*- coding: utf-8 -*-
"""
    jinja2.optimizer
    ~~~~~~~~~~~~~~~~

    The jinja optimizer is currently trying to constant fold a few expressions
    and modify the AST in place so that it should be easier to evaluate it.

    Because the AST does not contain all the scoping information and the
    compiler has to find that out, we cannot do all the optimizations we
    want.  For example loop unrolling doesn't work because unrolled loops would
    have a different scoping.

    The solution would be a second syntax tree that has the scoping rules stored.

    :copyright: (c) 2017 by the Jinja Team.
    :license: BSD.
"""
from pi._requires.jinja2 import nodes
from pi._requires.jinja2.visitor import NodeTransformer


def optimize(node, environment):
    """The context hint can be used to perform an static optimization
    based on the context given."""
    optimizer = Optimizer(environment)
    return optimizer.visit(node)


class Optimizer(NodeTransformer):

    def __init__(self, environment):
        self.environment = environment

    def fold(self, node, eval_ctx=None):
        """Do constant folding."""
        node = self.generic_visit(node)
        try:
            return nodes.Const.from_untrusted(node.as_const(eval_ctx),
                                              lineno=node.lineno,
                                              environment=self.environment)
        except nodes.Impossible:
            return node

    visit_Add = visit_Sub = visit_Mul = visit_Div = visit_FloorDiv = \
    visit_Pow = visit_Mod = visit_And = visit_Or = visit_Pos = visit_Neg = \
    visit_Not = visit_Compare = visit_Getitem = visit_Getattr = visit_Call = \
    visit_Filter = visit_Test = visit_CondExpr = fold
    del fold



import errno
import os
import select
import struct

import pi._requires.six

try:
    from ..transport import NpipeSocket
except ImportError:
    NpipeSocket = type(None)


class SocketError(Exception):
    pass


def read(socket, n=4096):
    """
    Reads at most n bytes from socket
    """

    recoverable_errors = (errno.EINTR, errno.EDEADLK, errno.EWOULDBLOCK)

    # wait for data to become available
    if not isinstance(socket, NpipeSocket):
        select.select([socket], [], [])

    try:
        if hasattr(socket, 'recv'):
            return socket.recv(n)
        return os.read(socket.fileno(), n)
    except EnvironmentError as e:
        if e.errno not in recoverable_errors:
            raise


def read_exactly(socket, n):
    """
    Reads exactly n bytes from socket
    Raises SocketError if there isn't enough data
    """
    data = pi._requires.six.binary_type()
    while len(data) < n:
        next_data = read(socket, n - len(data))
        if not next_data:
            raise SocketError("Unexpected EOF")
        data += next_data
    return data


def next_frame_size(socket):
    """
    Returns the size of the next frame of data waiting to be read from socket,
    according to the protocol defined here:

    https://docs.docker.com/engine/reference/api/docker_remote_api_v1.24/#/attach-to-a-container
    """
    try:
        data = read_exactly(socket, 8)
    except SocketError:
        return 0

    _, actual = struct.unpack('>BxxxL', data)
    return actual


def frames_iter(socket):
    """
    Returns a generator of frames read from socket
    """
    while True:
        n = next_frame_size(socket)
        if n == 0:
            break
        while n > 0:
            result = read(socket, n)
            n -= len(result)
            yield result



# Compatibility module. See https://github.com/docker/docker-py/issues/1196

import warnings

from ..types import Ulimit, LogConfig  # flake8: noqa

warnings.warn('docker.utils.types is now docker.types', ImportWarning)



import logging

from .._requires import click

from ..types import Meta
from ..config import read_config
from ..images import get_images
from ..environ import Environ
from ..console import configure_logging
from ..services import get_services

from .image import create_images_cli
from .custom import create_commands_cli
from .service import create_service_cli


log = logging.getLogger(__name__)


class UI(click.CommandCollection):

    def __init__(self, config, meta, core, custom, **kwargs):
        params = [click.Option(['--debug'], is_flag=True,
                               help='Run in debug mode')]
        super().__init__(sources=(core + custom), callback=self.callback,
                         params=params, help=meta.description, **kwargs)
        self._meta = meta
        self._config = config
        self._core = core
        self._custom = custom

    def callback(self, debug):
        configure_logging(debug)

        ctx = click.get_current_context()
        images = get_images(self._config)
        services = get_services(self._config)
        ctx.obj = Environ(self._meta, images, services)
        log.debug('Environment configured')

    def _list_core_commands(self, ctx):
        rv = set()
        for source in self._core:
            rv.update(source.list_commands(ctx))
        return sorted(rv)

    def _list_custom_commands(self, ctx):
        rv = set()
        for source in self._custom:
            rv.update(source.list_commands(ctx))
        return sorted(rv)

    def format_commands(self, ctx, formatter):

        def get_rows(commands):
            click_commands = [self.get_command(ctx, c) for c in commands]
            rows = []
            if any(isinstance(c, click.MultiCommand) for c in click_commands):
                for cmd in click_commands:
                    prefix = '+' if isinstance(cmd, click.MultiCommand) else ' '
                    rows.append((prefix + ' ' + cmd.name, cmd.short_help or ''))
            else:
                for cmd in click_commands:
                    rows.append((cmd.name, cmd.short_help or ''))
            return rows

        core = get_rows(self._list_core_commands(ctx))
        custom = get_rows(self._list_custom_commands(ctx))
        if core:
            with formatter.section('Core commands'):
                formatter.write_dl(core)
        if custom:
            with formatter.section('Custom commands'):
                formatter.write_dl(custom)


def build_ui():
    config = read_config()

    meta = Meta()
    for obj in config:
        if isinstance(obj, Meta):
            meta = obj

    images_cli = create_images_cli()
    services_cli = create_service_cli()
    commands_cli = create_commands_cli(config)

    return UI(config, meta, [images_cli, services_cli], [commands_cli])



from .._requires import click


class ExtGroup(click.Group):

    def __init__(self, *args, ext_help=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.ext_help = ext_help

    def parse_args(self, ctx, args):
        if not args:
            click.echo(ctx.get_help(), color=ctx.color)
            ctx.exit()
        return super().parse_args(ctx, args)

    def format_help_text(self, ctx, formatter):
        super().format_help_text(ctx, formatter)
        if self.ext_help is not None:
            self.ext_help(ctx, formatter)


class ProxyCommand(click.Command):

    def parse_args(self, ctx, args):
        ctx.args = args

    def invoke(self, ctx):
        if self.callback is not None:
            ctx.invoke(self.callback, args=ctx.args)



# -*- coding: utf-8 -*-
"""
    jinja2.asyncsupport
    ~~~~~~~~~~~~~~~~~~~

    Has all the code for async support which is implemented as a patch
    for supported Python versions.

    :copyright: (c) 2017 by the Jinja Team.
    :license: BSD, see LICENSE for more details.
"""
import sys
import asyncio
import inspect
from functools import update_wrapper

from pi._requires.jinja2.utils import concat, internalcode, Markup
from pi._requires.jinja2.environment import TemplateModule
from pi._requires.jinja2.runtime import LoopContextBase, _last_iteration


async def concat_async(async_gen):
    rv = []
    async def collect():
        async for event in async_gen:
            rv.append(event)
    await collect()
    return concat(rv)


async def generate_async(self, *args, **kwargs):
    vars = dict(*args, **kwargs)
    try:
        async for event in self.root_render_func(self.new_context(vars)):
            yield event
    except Exception:
        exc_info = sys.exc_info()
    else:
        return
    yield self.environment.handle_exception(exc_info, True)


def wrap_generate_func(original_generate):
    def _convert_generator(self, loop, args, kwargs):
        async_gen = self.generate_async(*args, **kwargs)
        try:
            while 1:
                yield loop.run_until_complete(async_gen.__anext__())
        except StopAsyncIteration:
            pass
    def generate(self, *args, **kwargs):
        if not self.environment.is_async:
            return original_generate(self, *args, **kwargs)
        return _convert_generator(self, asyncio.get_event_loop(), args, kwargs)
    return update_wrapper(generate, original_generate)


async def render_async(self, *args, **kwargs):
    if not self.environment.is_async:
        raise RuntimeError('The environment was not created with async mode '
                           'enabled.')

    vars = dict(*args, **kwargs)
    ctx = self.new_context(vars)

    try:
        return await concat_async(self.root_render_func(ctx))
    except Exception:
        exc_info = sys.exc_info()
    return self.environment.handle_exception(exc_info, True)


def wrap_render_func(original_render):
    def render(self, *args, **kwargs):
        if not self.environment.is_async:
            return original_render(self, *args, **kwargs)
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.render_async(*args, **kwargs))
    return update_wrapper(render, original_render)


def wrap_block_reference_call(original_call):
    @internalcode
    async def async_call(self):
        rv = await concat_async(self._stack[self._depth](self._context))
        if self._context.eval_ctx.autoescape:
            rv = Markup(rv)
        return rv

    @internalcode
    def __call__(self):
        if not self._context.environment.is_async:
            return original_call(self)
        return async_call(self)

    return update_wrapper(__call__, original_call)


def wrap_macro_invoke(original_invoke):
    @internalcode
    async def async_invoke(self, arguments, autoescape):
        rv = await self._func(*arguments)
        if autoescape:
            rv = Markup(rv)
        return rv

    @internalcode
    def _invoke(self, arguments, autoescape):
        if not self._environment.is_async:
            return original_invoke(self, arguments, autoescape)
        return async_invoke(self, arguments, autoescape)
    return update_wrapper(_invoke, original_invoke)


@internalcode
async def get_default_module_async(self):
    if self._module is not None:
        return self._module
    self._module = rv = await self.make_module_async()
    return rv


def wrap_default_module(original_default_module):
    @internalcode
    def _get_default_module(self):
        if self.environment.is_async:
            raise RuntimeError('Template module attribute is unavailable '
                               'in async mode')
        return original_default_module(self)
    return _get_default_module


async def make_module_async(self, vars=None, shared=False, locals=None):
    context = self.new_context(vars, shared, locals)
    body_stream = []
    async for item in self.root_render_func(context):
        body_stream.append(item)
    return TemplateModule(self, context, body_stream)


def patch_template():
    from pi._requires.jinja2 import Template
    Template.generate = wrap_generate_func(Template.generate)
    Template.generate_async = update_wrapper(
        generate_async, Template.generate_async)
    Template.render_async = update_wrapper(
        render_async, Template.render_async)
    Template.render = wrap_render_func(Template.render)
    Template._get_default_module = wrap_default_module(
        Template._get_default_module)
    Template._get_default_module_async = get_default_module_async
    Template.make_module_async = update_wrapper(
        make_module_async, Template.make_module_async)


def patch_runtime():
    from pi._requires.jinja2.runtime import BlockReference, Macro
    BlockReference.__call__ = wrap_block_reference_call(
        BlockReference.__call__)
    Macro._invoke = wrap_macro_invoke(Macro._invoke)


def patch_filters():
    from pi._requires.jinja2.filters import FILTERS
    from pi._requires.jinja2.asyncfilters import ASYNC_FILTERS
    FILTERS.update(ASYNC_FILTERS)


def patch_all():
    patch_template()
    patch_runtime()
    patch_filters()


async def auto_await(value):
    if inspect.isawaitable(value):
        return await value
    return value


async def auto_aiter(iterable):
    if hasattr(iterable, '__aiter__'):
        async for item in iterable:
            yield item
        return
    for item in iterable:
        yield item


class AsyncLoopContext(LoopContextBase):

    def __init__(self, async_iterator, after, length, recurse=None,
                 depth0=0):
        LoopContextBase.__init__(self, recurse, depth0)
        self._async_iterator = async_iterator
        self._after = after
        self._length = length

    @property
    def length(self):
        if self._length is None:
            raise TypeError('Loop length for some iterators cannot be '
                            'lazily calculated in async mode')
        return self._length

    def __aiter__(self):
        return AsyncLoopContextIterator(self)


class AsyncLoopContextIterator(object):
    __slots__ = ('context',)

    def __init__(self, context):
        self.context = context

    def __aiter__(self):
        return self

    async def __anext__(self):
        ctx = self.context
        ctx.index0 += 1
        if ctx._after is _last_iteration:
            raise StopAsyncIteration()
        next_elem = ctx._after
        try:
            ctx._after = await ctx._async_iterator.__anext__()
        except StopAsyncIteration:
            ctx._after = _last_iteration
        return next_elem, ctx


async def make_async_loop_context(iterable, recurse=None, depth0=0):
    # Length is more complicated and less efficient in async mode.  The
    # reason for this is that we cannot know if length will be used
    # upfront but because length is a property we cannot lazily execute it
    # later.  This means that we need to buffer it up and measure :(
    #
    # We however only do this for actual iterators, not for async
    # iterators as blocking here does not seem like the best idea in the
    # world.
    try:
        length = len(iterable)
    except (TypeError, AttributeError):
        if not hasattr(iterable, '__aiter__'):
            iterable = tuple(iterable)
            length = len(iterable)
        else:
            length = None
    async_iterator = auto_aiter(iterable)
    try:
        after = await async_iterator.__anext__()
    except StopAsyncIteration:
        after = _last_iteration
    return AsyncLoopContext(async_iterator, after, length, recurse, depth0)



# -*- coding: utf-8 -*-
"""
    markupsafe._native
    ~~~~~~~~~~~~~~~~~~

    Native Python implementation the C module is not compiled.

    :copyright: (c) 2010 by Armin Ronacher.
    :license: BSD, see LICENSE for more details.
"""
from pi._requires.markupsafe import Markup
from pi._requires.markupsafe._compat import text_type


def escape(s):
    """Convert the characters &, <, >, ' and " in string s to HTML-safe
    sequences.  Use this if you need to display text that might contain
    such characters in HTML.  Marks return value as markup string.
    """
    if hasattr(s, '__html__'):
        return s.__html__()
    return Markup(text_type(s)
        .replace('&', '&amp;')
        .replace('>', '&gt;')
        .replace('<', '&lt;')
        .replace("'", '&#39;')
        .replace('"', '&#34;')
    )


def escape_silent(s):
    """Like :func:`escape` but converts `None` into an empty
    markup string.
    """
    if s is None:
        return Markup()
    return escape(s)


def soft_unicode(s):
    """Make a string unicode if it isn't already.  That way a markup
    string is not converted back to unicode.
    """
    if not isinstance(s, text_type):
        s = text_type(s)
    return s



import sys
import signal
import socket
import os.path
import logging

from asyncio import Queue, CancelledError, Event
from asyncio import TimeoutError as AIOTimeoutError, wait_for

from ._requires import click

from .utils import terminate
from .client import APIError, NotFound


log = logging.getLogger(__name__)


async def stdin_reader(fd, in_queue, *, loop):
    eof = Event(loop=loop)

    def cb():
        data = os.read(fd, 32)
        if not data:
            eof.set()
        in_queue.put_nowait(data)

    loop.add_reader(fd, cb)
    try:
        await eof.wait()
    finally:
        loop.remove_reader(fd)


async def stdout_writer(out_queue):
    while True:
        data = await out_queue.get()
        sys.stdout.write(data.decode('utf-8', 'replace'))
        sys.stdout.flush()


async def socket_reader(sock, out_queue, *, loop):
    while True:
        data = await loop.sock_recv(sock, 4096)
        if not data:
            break
        await out_queue.put(data)


async def socket_writer(sock, in_queue, *, loop):
    while True:
        data = await in_queue.get()
        await loop.sock_sendall(sock, data)


class _VolumeBinds:

    @classmethod
    def translate_binds(cls, volumes):
        self = cls()
        return dict(self.visit(vol) for vol in volumes)

    @classmethod
    def translate_volumes(cls, volumes):
        return [os.path.abspath(vol.to) for vol in volumes]

    def visit(self, obj):
        return obj.accept(self)

    def visit_RO(self, obj):
        return 'ro'

    def visit_RW(self, obj):
        return 'rw'

    def visit_localpath(self, obj):
        from_ = os.path.abspath(obj.from_)
        to = os.path.abspath(obj.to)
        # FIXME: implement proper errors reporting
        assert os.path.exists(from_),\
            'Local path does not exists: {}'.format(from_)
        return from_, {'bind': to, 'mode': self.visit(obj.mode)}

    def visit_namedvolume(self, obj):
        to = os.path.abspath(obj.to)
        return obj.name, {'bind': to, 'mode': self.visit(obj.mode)}


def _port_binds(ports):
    return {'{}/{}'.format(e.port, e.proto): {'HostPort': e.as_,
                                              'HostIp': e.addr}
            for e in ports}


async def start(client, image, command, *, init=None, tty=True,
                entrypoint=None, volumes=None, ports=None, environ=None,
                work_dir=None, network=None, network_alias=None, label=None):
    volumes = volumes or []
    container_volumes = _VolumeBinds.translate_volumes(volumes)
    container_volume_binds = _VolumeBinds.translate_binds(volumes)

    ports = ports or []
    container_ports = [(e.port, e.proto) for e in ports]
    container_port_binds = _port_binds(ports)

    work_dir = os.path.abspath(work_dir) if work_dir is not None else None
    labels = [label] if label is not None else []

    host_config = client.create_host_config(
        init=init,
        binds=container_volume_binds,
        port_bindings=container_port_binds,
        network_mode=network,
    )
    networking_config = None
    if network_alias is not None:
        networking_config = client.create_networking_config(endpoints_config={
            network: {'Aliases': [network_alias]},
        })
    try:
        c = await client.create_container(
            image=image.name,
            command=command,
            stdin_open=True,
            tty=tty,
            ports=container_ports,
            environment=environ,
            volumes=container_volumes,
            entrypoint=entrypoint,
            working_dir=work_dir,
            labels=labels,
            host_config=host_config,
            networking_config=networking_config,
        )
    except APIError as e:
        click.echo(e.explanation)
        return
    try:
        await client.start(c)
    except APIError as e:
        click.echo(e.explanation)
        await client.remove_container(c, v=True, force=True)
    else:
        return c


async def resize(client, container):
    width, height = click.get_terminal_size()
    try:
        await client.resize(container, height, width)
    except NotFound as e:
        log.debug('Failed to resize terminal: %s', e)


async def attach(client, container, stdin_fd, *, loop, wait_exit=10):
    params = {'logs': 1, 'stdin': 1, 'stdout': 1, 'stderr': 1, 'stream': 1}
    async with client.attach_socket(container, params) as sock_io:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM,
                             fileno=sock_io.fileno())
        sock.setblocking(False)

        container_input = Queue(loop=loop)
        socket_writer_task = loop.create_task(
            socket_writer(sock, container_input, loop=loop))
        stdin_reader_task = loop.create_task(
            stdin_reader(stdin_fd, container_input, loop=loop))

        container_output = Queue(loop=loop)
        stdout_writer_task = loop.create_task(
            stdout_writer(container_output))
        socket_reader_task = loop.create_task(
            socket_reader(sock, container_output, loop=loop))

        exit_code = None
        try:
            exit_code = await client.wait(container)
        except CancelledError:
            await client.kill(container, signal.SIGINT)
            try:
                await wait_for(socket_reader_task, wait_exit, loop=loop)
            except AIOTimeoutError:
                await client.kill(container, signal.SIGKILL)

        await terminate(stdout_writer_task, loop=loop)
        await terminate(stdin_reader_task, loop=loop)
        await terminate(socket_reader_task, loop=loop)
        await terminate(socket_writer_task, loop=loop)
        return exit_code


async def run(client, stdin_fd, tty, image, command, *, loop, init=None,
              volumes=None, ports=None, environ=None, work_dir=None,
              network=None, network_alias=None, wait_exit=3):
    c = await start(client, image, command, init=init, tty=tty, volumes=volumes,
                    ports=ports, environ=environ, work_dir=work_dir,
                    network=network, network_alias=network_alias)
    if c is None:
        return
    try:
        await resize(client, c)
        exit_code = await attach(client, c, stdin_fd, loop=loop,
                                 wait_exit=wait_exit)
        if exit_code is None:
            exit_code = await client.wait(c)
        return exit_code

    finally:
        await client.remove_container(c, v=True, force=True)



import asyncio

from .utils import cached_property, SequenceMap, async_func


class Environ:

    def __init__(self, meta, images, services):
        self._meta = meta
        self.loop = asyncio.get_event_loop()
        self.images = SequenceMap(images, lambda i: i.name)
        self.services = SequenceMap(services, lambda i: i.name)

    @property
    def namespace(self):
        return self._meta.namespace or 'pi'

    @property
    def network(self):
        if self._meta.namespace:
            return 'pi-{}'.format(self._meta.namespace)
        else:
            return 'pi'

    @cached_property
    def client(self):
        from .client import AsyncClient

        return AsyncClient(loop=self.loop)


def async_cmd(func):

    @async_func()
    async def async_wrapper(env, *args, loop, **kwargs):
        await func(env, *args, **kwargs)

    def sync_wrapper(env, *args, **kwargs):
        async_wrapper(env, *args, loop=env.loop, **kwargs)

    return sync_wrapper



import sys
import hashlib

from .types import DockerImage, Image, ActionType


class Hasher:

    def visit(self, obj):
        return obj.accept(self)

    def visit_image(self, obj):
        yield obj.repository.encode('utf-8')
        for task in obj.tasks:
            yield from self.visit(task)

    def visit_task(self, obj):
        yield obj.run.encode('utf-8')
        for value in obj.where.values():
            if isinstance(value, ActionType):
                yield from self.visit(value)
            else:
                yield str(value).encode('utf-8')

    def visit_download(self, obj):
        yield obj.url.encode('utf-8')

    def visit_file(self, obj):
        with open(obj.path, 'rb') as f:
            while True:
                chunk = f.read(2**16)
                if not chunk:
                    break
                else:
                    yield chunk

    def visit_bundle(self, obj):
        yield obj.path.encode('utf-8')


def image_hashes(images_map, images, *, _cache=None):
    _cache = _cache or {}

    hasher = Hasher()
    hashes = []
    for image in images:
        if image.name in _cache:
            hashes.append(_cache[image.name])
            continue

        if isinstance(image.from_, DockerImage):
            parent_hashable = image.from_.name
        else:
            parent = images_map.get(image.from_)
            parent_hashable, = image_hashes(images_map, [parent],
                                            _cache=_cache)

        h = hashlib.sha1()
        h.update(parent_hashable.encode('utf-8'))
        for chunk in hasher.visit(image):
            h.update(chunk)
        hex_digest = _cache[image.name] = h.hexdigest()
        hashes.append(hex_digest)
    return hashes


def image_versions(images_map, images):
    hashes = image_hashes(images_map, images)
    return [h[:12] for h in hashes]


def docker_image(images_map, image):
    if isinstance(image, str):
        image = images_map.get(image)
        version, = image_versions(images_map, [image])
        return DockerImage.from_image(image, version)
    elif isinstance(image, DockerImage):
        return image
    else:
        raise TypeError(repr(image))


def resolve_deps(deps):
    while deps:
        resolved = set()
        for name, parent_name in deps.items():
            if parent_name not in deps:
                resolved.add(name)
        if not resolved:
            raise TypeError('Images hierarchy build error, '
                            'circular dependency found in these images: {}'
                            .format(', '.join(sorted(deps.keys()))))
        for name in resolved:
            yield name, deps[name]
        deps = {k: v for k, v in deps.items() if k not in resolved}


def get_images(config):
    return [i for i in config if isinstance(i, Image)]


async def _echo_download_progress(output):
    error = False
    last_id = None
    while True:
        items = await output.read()
        if not items:
            break
        for i in items:
            error = error or 'error' in i

            progress_id = i.get('id')
            if last_id:
                if progress_id == last_id:
                    sys.stdout.write('\x1b[2K\r')
                elif not progress_id or progress_id != last_id:
                    sys.stdout.write('\n')
            last_id = progress_id

            if progress_id:
                sys.stdout.write('{}: '.format(progress_id))
            sys.stdout.write(i.get('status') or i.get('error') or '')

            progress_bar = i.get('progress')
            if progress_bar:
                sys.stdout.write(' ' + progress_bar)

            if not progress_id:
                sys.stdout.write('\n')
            sys.stdout.flush()
    if last_id:
        sys.stdout.write('\n')
        sys.stdout.flush()
    return not error


async def pull(client, docker_image):
    from .client import APIError

    try:
        output = await client.pull(docker_image.name, stream=True, decode=True)
    except APIError as e:
        if e.response.status_code == 404:
            return False
        raise
    else:
        with output as reader:
            success = await _echo_download_progress(reader)
            return success


async def push(client, docker_image):
    output = await client.push(docker_image.name, stream=True, decode=True)
    with output as reader:
        success = await _echo_download_progress(reader)
        return success


async def build(client, images_map, image, *, loop):
    from .tasks import build as build_tasks

    result = await build_tasks(client, images_map, image, loop=loop)
    return result



from __future__ import absolute_import, division, print_function


class FrozenInstanceError(AttributeError):
    """
    A frozen/immutable instance has been attempted to be modified.

    It mirrors the behavior of ``namedtuples`` by using the same error message
    and subclassing :exc:`AttributeError`.

    .. versionadded:: 16.1.0
    """
    msg = "can't set attribute"
    args = [msg]


class AttrsAttributeNotFoundError(ValueError):
    """
    An ``attrs`` function couldn't find an attribute that the user asked for.

    .. versionadded:: 16.2.0
    """


class NotAnAttrsClassError(ValueError):
    """
    A non-``attrs`` class has been passed into an ``attrs`` function.

    .. versionadded:: 16.2.0
    """



import copy

from ..api import APIClient
from ..errors import (ContainerError, ImageNotFound,
                      create_unexpected_kwargs_error)
from ..types import HostConfig
from .images import Image
from .resource import Collection, Model


class Container(Model):

    @property
    def name(self):
        """
        The name of the container.
        """
        if self.attrs.get('Name') is not None:
            return self.attrs['Name'].lstrip('/')

    @property
    def image(self):
        """
        The image of the container.
        """
        image_id = self.attrs['Image']
        if image_id is None:
            return None
        return self.client.images.get(image_id.split(':')[1])

    @property
    def labels(self):
        """
        The labels of a container as dictionary.
        """
        result = self.attrs['Config'].get('Labels')
        return result or {}

    @property
    def status(self):
        """
        The status of the container. For example, ``running``, or ``exited``.
        """
        return self.attrs['State']['Status']

    def attach(self, **kwargs):
        """
        Attach to this container.

        :py:meth:`logs` is a wrapper around this method, which you can
        use instead if you want to fetch/stream container output without first
        retrieving the entire backlog.

        Args:
            stdout (bool): Include stdout.
            stderr (bool): Include stderr.
            stream (bool): Return container output progressively as an iterator
                of strings, rather than a single string.
            logs (bool): Include the container's previous output.

        Returns:
            By default, the container's output as a single string.

            If ``stream=True``, an iterator of output strings.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.attach(self.id, **kwargs)

    def attach_socket(self, **kwargs):
        """
        Like :py:meth:`attach`, but returns the underlying socket-like object
        for the HTTP request.

        Args:
            params (dict): Dictionary of request parameters (e.g. ``stdout``,
                ``stderr``, ``stream``).
            ws (bool): Use websockets instead of raw HTTP.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.attach_socket(self.id, **kwargs)

    def commit(self, repository=None, tag=None, **kwargs):
        """
        Commit a container to an image. Similar to the ``docker commit``
        command.

        Args:
            repository (str): The repository to push the image to
            tag (str): The tag to push
            message (str): A commit message
            author (str): The name of the author
            changes (str): Dockerfile instructions to apply while committing
            conf (dict): The configuration for the container. See the
                `Engine API documentation
                <https://docs.docker.com/reference/api/docker_remote_api/>`_
                for full details.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """

        resp = self.client.api.commit(self.id, repository=repository, tag=tag,
                                      **kwargs)
        return self.client.images.get(resp['Id'])

    def diff(self):
        """
        Inspect changes on a container's filesystem.

        Returns:
            (str)

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.diff(self.id)

    def exec_run(self, cmd, stdout=True, stderr=True, stdin=False, tty=False,
                 privileged=False, user='', detach=False, stream=False,
                 socket=False, environment=None):
        """
        Run a command inside this container. Similar to
        ``docker exec``.

        Args:
            cmd (str or list): Command to be executed
            stdout (bool): Attach to stdout. Default: ``True``
            stderr (bool): Attach to stderr. Default: ``True``
            stdin (bool): Attach to stdin. Default: ``False``
            tty (bool): Allocate a pseudo-TTY. Default: False
            privileged (bool): Run as privileged.
            user (str): User to execute command as. Default: root
            detach (bool): If true, detach from the exec command.
                Default: False
            stream (bool): Stream response data. Default: False
            environment (dict or list): A dictionary or a list of strings in
                the following format ``["PASSWORD=xxx"]`` or
                ``{"PASSWORD": "xxx"}``.

        Returns:
            (generator or str): If ``stream=True``, a generator yielding
                response chunks. A string containing response data otherwise.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        resp = self.client.api.exec_create(
            self.id, cmd, stdout=stdout, stderr=stderr, stdin=stdin, tty=tty,
            privileged=privileged, user=user, environment=environment
        )
        return self.client.api.exec_start(
            resp['Id'], detach=detach, tty=tty, stream=stream, socket=socket
        )

    def export(self):
        """
        Export the contents of the container's filesystem as a tar archive.

        Returns:
            (str): The filesystem tar archive

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.export(self.id)

    def get_archive(self, path):
        """
        Retrieve a file or folder from the container in the form of a tar
        archive.

        Args:
            path (str): Path to the file or folder to retrieve

        Returns:
            (tuple): First element is a raw tar data stream. Second element is
            a dict containing ``stat`` information on the specified ``path``.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.get_archive(self.id, path)

    def kill(self, signal=None):
        """
        Kill or send a signal to the container.

        Args:
            signal (str or int): The signal to send. Defaults to ``SIGKILL``

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """

        return self.client.api.kill(self.id, signal=signal)

    def logs(self, **kwargs):
        """
        Get logs from this container. Similar to the ``docker logs`` command.

        The ``stream`` parameter makes the ``logs`` function return a blocking
        generator you can iterate over to retrieve log output as it happens.

        Args:
            stdout (bool): Get ``STDOUT``
            stderr (bool): Get ``STDERR``
            stream (bool): Stream the response
            timestamps (bool): Show timestamps
            tail (str or int): Output specified number of lines at the end of
                logs. Either an integer of number of lines or the string
                ``all``. Default ``all``
            since (datetime or int): Show logs since a given datetime or
                integer epoch (in seconds)
            follow (bool): Follow log output

        Returns:
            (generator or str): Logs from the container.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.logs(self.id, **kwargs)

    def pause(self):
        """
        Pauses all processes within this container.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.pause(self.id)

    def put_archive(self, path, data):
        """
        Insert a file or folder in this container using a tar archive as
        source.

        Args:
            path (str): Path inside the container where the file(s) will be
                extracted. Must exist.
            data (bytes): tar data to be extracted

        Returns:
            (bool): True if the call succeeds.

        Raises:
            :py:class:`~docker.errors.APIError` If an error occurs.
        """
        return self.client.api.put_archive(self.id, path, data)

    def remove(self, **kwargs):
        """
        Remove this container. Similar to the ``docker rm`` command.

        Args:
            v (bool): Remove the volumes associated with the container
            link (bool): Remove the specified link and not the underlying
                container
            force (bool): Force the removal of a running container (uses
                ``SIGKILL``)

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.remove_container(self.id, **kwargs)

    def rename(self, name):
        """
        Rename this container. Similar to the ``docker rename`` command.

        Args:
            name (str): New name for the container

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.rename(self.id, name)

    def resize(self, height, width):
        """
        Resize the tty session.

        Args:
            height (int): Height of tty session
            width (int): Width of tty session

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.resize(self.id, height, width)

    def restart(self, **kwargs):
        """
        Restart this container. Similar to the ``docker restart`` command.

        Args:
            timeout (int): Number of seconds to try to stop for before killing
                the container. Once killed it will then be restarted. Default
                is 10 seconds.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.restart(self.id, **kwargs)

    def start(self, **kwargs):
        """
        Start this container. Similar to the ``docker start`` command, but
        doesn't support attach options.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.start(self.id, **kwargs)

    def stats(self, **kwargs):
        """
        Stream statistics for this container. Similar to the
        ``docker stats`` command.

        Args:
            decode (bool): If set to true, stream will be decoded into dicts
                on the fly. False by default.
            stream (bool): If set to false, only the current stats will be
                returned instead of a stream. True by default.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.stats(self.id, **kwargs)

    def stop(self, **kwargs):
        """
        Stops a container. Similar to the ``docker stop`` command.

        Args:
            timeout (int): Timeout in seconds to wait for the container to
                stop before sending a ``SIGKILL``. Default: 10

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.stop(self.id, **kwargs)

    def top(self, **kwargs):
        """
        Display the running processes of the container.

        Args:
            ps_args (str): An optional arguments passed to ps (e.g. ``aux``)

        Returns:
            (str): The output of the top

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.top(self.id, **kwargs)

    def unpause(self):
        """
        Unpause all processes within the container.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.unpause(self.id)

    def update(self, **kwargs):
        """
        Update resource configuration of the containers.

        Args:
            blkio_weight (int): Block IO (relative weight), between 10 and 1000
            cpu_period (int): Limit CPU CFS (Completely Fair Scheduler) period
            cpu_quota (int): Limit CPU CFS (Completely Fair Scheduler) quota
            cpu_shares (int): CPU shares (relative weight)
            cpuset_cpus (str): CPUs in which to allow execution
            cpuset_mems (str): MEMs in which to allow execution
            mem_limit (int or str): Memory limit
            mem_reservation (int or str): Memory soft limit
            memswap_limit (int or str): Total memory (memory + swap), -1 to
                disable swap
            kernel_memory (int or str): Kernel memory limit
            restart_policy (dict): Restart policy dictionary

        Returns:
            (dict): Dictionary containing a ``Warnings`` key.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.update_container(self.id, **kwargs)

    def wait(self, **kwargs):
        """
        Block until the container stops, then return its exit code. Similar to
        the ``docker wait`` command.

        Args:
            timeout (int): Request timeout

        Returns:
            (int): The exit code of the container. Returns ``-1`` if the API
            responds without a ``StatusCode`` attribute.

        Raises:
            :py:class:`requests.exceptions.ReadTimeout`
                If the timeout is exceeded.
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.wait(self.id, **kwargs)


class ContainerCollection(Collection):
    model = Container

    def run(self, image, command=None, stdout=True, stderr=False,
            remove=False, **kwargs):
        """
        Run a container. By default, it will wait for the container to finish
        and return its logs, similar to ``docker run``.

        If the ``detach`` argument is ``True``, it will start the container
        and immediately return a :py:class:`Container` object, similar to
        ``docker run -d``.

        Example:
            Run a container and get its output:

            >>> import docker
            >>> client = docker.from_env()
            >>> client.containers.run('alpine', 'echo hello world')
            b'hello world\\n'

            Run a container and detach:

            >>> container = client.containers.run('bfirsh/reticulate-splines',
                                                  detach=True)
            >>> container.logs()
            'Reticulating spline 1...\\nReticulating spline 2...\\n'

        Args:
            image (str): The image to run.
            command (str or list): The command to run in the container.
            auto_remove (bool): enable auto-removal of the container on daemon
                side when the container's process exits.
            blkio_weight_device: Block IO weight (relative device weight) in
                the form of: ``[{"Path": "device_path", "Weight": weight}]``.
            blkio_weight: Block IO weight (relative weight), accepts a weight
                value between 10 and 1000.
            cap_add (list of str): Add kernel capabilities. For example,
                ``["SYS_ADMIN", "MKNOD"]``.
            cap_drop (list of str): Drop kernel capabilities.
            cpu_count (int): Number of usable CPUs (Windows only).
            cpu_percent (int): Usable percentage of the available CPUs
                (Windows only).
            cpu_period (int): The length of a CPU period in microseconds.
            cpu_quota (int): Microseconds of CPU time that the container can
                get in a CPU period.
            cpu_shares (int): CPU shares (relative weight).
            cpuset_cpus (str): CPUs in which to allow execution (``0-3``,
                ``0,1``).
            cpuset_mems (str): Memory nodes (MEMs) in which to allow execution
                (``0-3``, ``0,1``). Only effective on NUMA systems.
            detach (bool): Run container in the background and return a
                :py:class:`Container` object.
            device_read_bps: Limit read rate (bytes per second) from a device
                in the form of: `[{"Path": "device_path", "Rate": rate}]`
            device_read_iops: Limit read rate (IO per second) from a device.
            device_write_bps: Limit write rate (bytes per second) from a
                device.
            device_write_iops: Limit write rate (IO per second) from a device.
            devices (:py:class:`list`): Expose host devices to the container,
                as a list of strings in the form
                ``<path_on_host>:<path_in_container>:<cgroup_permissions>``.

                For example, ``/dev/sda:/dev/xvda:rwm`` allows the container
                to have read-write access to the host's ``/dev/sda`` via a
                node named ``/dev/xvda`` inside the container.
            dns (:py:class:`list`): Set custom DNS servers.
            dns_opt (:py:class:`list`): Additional options to be added to the
                container's ``resolv.conf`` file.
            dns_search (:py:class:`list`): DNS search domains.
            domainname (str or list): Set custom DNS search domains.
            entrypoint (str or list): The entrypoint for the container.
            environment (dict or list): Environment variables to set inside
                the container, as a dictionary or a list of strings in the
                format ``["SOMEVARIABLE=xxx"]``.
            extra_hosts (dict): Addtional hostnames to resolve inside the
                container, as a mapping of hostname to IP address.
            group_add (:py:class:`list`): List of additional group names and/or
                IDs that the container process will run as.
            hostname (str): Optional hostname for the container.
            init (bool): Run an init inside the container that forwards
                signals and reaps processes
            init_path (str): Path to the docker-init binary
            ipc_mode (str): Set the IPC mode for the container.
            isolation (str): Isolation technology to use. Default: `None`.
            labels (dict or list): A dictionary of name-value labels (e.g.
                ``{"label1": "value1", "label2": "value2"}``) or a list of
                names of labels to set with empty values (e.g.
                ``["label1", "label2"]``)
            links (dict or list of tuples): Either a dictionary mapping name
                to alias or as a list of ``(name, alias)`` tuples.
            log_config (dict): Logging configuration, as a dictionary with
                keys:

                - ``type`` The logging driver name.
                - ``config`` A dictionary of configuration for the logging
                  driver.

            mac_address (str): MAC address to assign to the container.
            mem_limit (int or str): Memory limit. Accepts float values
                (which represent the memory limit of the created container in
                bytes) or a string with a units identification char
                (``100000b``, ``1000k``, ``128m``, ``1g``). If a string is
                specified without a units character, bytes are assumed as an
                intended unit.
            mem_swappiness (int): Tune a container's memory swappiness
                behavior. Accepts number between 0 and 100.
            memswap_limit (str or int): Maximum amount of memory + swap a
                container is allowed to consume.
            name (str): The name for this container.
            nano_cpus (int):  CPU quota in units of 10-9 CPUs.
            network (str): Name of the network this container will be connected
                to at creation time. You can connect to additional networks
                using :py:meth:`Network.connect`. Incompatible with
                ``network_mode``.
            network_disabled (bool): Disable networking.
            network_mode (str): One of:

                - ``bridge`` Create a new network stack for the container on
                  on the bridge network.
                - ``none`` No networking for this container.
                - ``container:<name|id>`` Reuse another container's network
                  stack.
                - ``host`` Use the host network stack.
                Incompatible with ``network``.
            oom_kill_disable (bool): Whether to disable OOM killer.
            oom_score_adj (int): An integer value containing the score given
                to the container in order to tune OOM killer preferences.
            pid_mode (str): If set to ``host``, use the host PID namespace
                inside the container.
            pids_limit (int): Tune a container's pids limit. Set ``-1`` for
                unlimited.
            ports (dict): Ports to bind inside the container.

                The keys of the dictionary are the ports to bind inside the
                container, either as an integer or a string in the form
                ``port/protocol``, where the protocol is either ``tcp`` or
                ``udp``.

                The values of the dictionary are the corresponding ports to
                open on the host, which can be either:

                - The port number, as an integer. For example,
                  ``{'2222/tcp': 3333}`` will expose port 2222 inside the
                  container as port 3333 on the host.
                - ``None``, to assign a random host port. For example,
                  ``{'2222/tcp': None}``.
                - A tuple of ``(address, port)`` if you want to specify the
                  host interface. For example,
                  ``{'1111/tcp': ('127.0.0.1', 1111)}``.
                - A list of integers, if you want to bind multiple host ports
                  to a single container port. For example,
                  ``{'1111/tcp': [1234, 4567]}``.

            privileged (bool): Give extended privileges to this container.
            publish_all_ports (bool): Publish all ports to the host.
            read_only (bool): Mount the container's root filesystem as read
                only.
            remove (bool): Remove the container when it has finished running.
                Default: ``False``.
            restart_policy (dict): Restart the container when it exits.
                Configured as a dictionary with keys:

                - ``Name`` One of ``on-failure``, or ``always``.
                - ``MaximumRetryCount`` Number of times to restart the
                  container on failure.

                For example:
                ``{"Name": "on-failure", "MaximumRetryCount": 5}``

            security_opt (:py:class:`list`): A list of string values to
                customize labels for MLS systems, such as SELinux.
            shm_size (str or int): Size of /dev/shm (e.g. ``1G``).
            stdin_open (bool): Keep ``STDIN`` open even if not attached.
            stdout (bool): Return logs from ``STDOUT`` when ``detach=False``.
                Default: ``True``.
            stderr (bool): Return logs from ``STDERR`` when ``detach=False``.
                Default: ``False``.
            stop_signal (str): The stop signal to use to stop the container
                (e.g. ``SIGINT``).
            storage_opt (dict): Storage driver options per container as a
                key-value mapping.
            sysctls (dict): Kernel parameters to set in the container.
            tmpfs (dict): Temporary filesystems to mount, as a dictionary
                mapping a path inside the container to options for that path.

                For example:

                .. code-block:: python

                    {
                        '/mnt/vol2': '',
                        '/mnt/vol1': 'size=3G,uid=1000'
                    }

            tty (bool): Allocate a pseudo-TTY.
            ulimits (:py:class:`list`): Ulimits to set inside the container, as
                a list of dicts.
            user (str or int): Username or UID to run commands as inside the
                container.
            userns_mode (str): Sets the user namespace mode for the container
                when user namespace remapping option is enabled. Supported
                values are: ``host``
            volume_driver (str): The name of a volume driver/plugin.
            volumes (dict or list): A dictionary to configure volumes mounted
                inside the container. The key is either the host path or a
                volume name, and the value is a dictionary with the keys:

                - ``bind`` The path to mount the volume inside the container
                - ``mode`` Either ``rw`` to mount the volume read/write, or
                  ``ro`` to mount it read-only.

                For example:

                .. code-block:: python

                    {'/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
                     '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}}

            volumes_from (:py:class:`list`): List of container names or IDs to
                get volumes from.
            working_dir (str): Path to the working directory.

        Returns:
            The container logs, either ``STDOUT``, ``STDERR``, or both,
            depending on the value of the ``stdout`` and ``stderr`` arguments.

            If ``detach`` is ``True``, a :py:class:`Container` object is
            returned instead.

        Raises:
            :py:class:`docker.errors.ContainerError`
                If the container exits with a non-zero exit code and
                ``detach`` is ``False``.
            :py:class:`docker.errors.ImageNotFound`
                If the specified image does not exist.
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        if isinstance(image, Image):
            image = image.id
        detach = kwargs.pop("detach", False)
        if detach and remove:
            raise RuntimeError("The options 'detach' and 'remove' cannot be "
                               "used together.")

        if kwargs.get('network') and kwargs.get('network_mode'):
            raise RuntimeError(
                'The options "network" and "network_mode" can not be used '
                'together.'
            )

        try:
            container = self.create(image=image, command=command,
                                    detach=detach, **kwargs)
        except ImageNotFound:
            self.client.images.pull(image)
            container = self.create(image=image, command=command,
                                    detach=detach, **kwargs)

        container.start()

        if detach:
            return container

        exit_status = container.wait()
        if exit_status != 0:
            stdout = False
            stderr = True
        out = container.logs(stdout=stdout, stderr=stderr)
        if remove:
            container.remove()
        if exit_status != 0:
            raise ContainerError(container, exit_status, command, image, out)
        return out

    def create(self, image, command=None, **kwargs):
        """
        Create a container without starting it. Similar to ``docker create``.

        Takes the same arguments as :py:meth:`run`, except for ``stdout``,
        ``stderr``, and ``remove``.

        Returns:
            A :py:class:`Container` object.

        Raises:
            :py:class:`docker.errors.ImageNotFound`
                If the specified image does not exist.
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        if isinstance(image, Image):
            image = image.id
        kwargs['image'] = image
        kwargs['command'] = command
        kwargs['version'] = self.client.api._version
        create_kwargs = _create_container_args(kwargs)
        resp = self.client.api.create_container(**create_kwargs)
        return self.get(resp['Id'])

    def get(self, container_id):
        """
        Get a container by name or ID.

        Args:
            container_id (str): Container name or ID.

        Returns:
            A :py:class:`Container` object.

        Raises:
            :py:class:`docker.errors.NotFound`
                If the container does not exist.
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        resp = self.client.api.inspect_container(container_id)
        return self.prepare_model(resp)

    def list(self, all=False, before=None, filters=None, limit=-1, since=None):
        """
        List containers. Similar to the ``docker ps`` command.

        Args:
            all (bool): Show all containers. Only running containers are shown
                by default
            since (str): Show only containers created since Id or Name, include
                non-running ones
            before (str): Show only container created before Id or Name,
                include non-running ones
            limit (int): Show `limit` last created containers, include
                non-running ones
            filters (dict): Filters to be processed on the image list.
                Available filters:

                - `exited` (int): Only containers with specified exit code
                - `status` (str): One of ``restarting``, ``running``,
                    ``paused``, ``exited``
                - `label` (str): format either ``"key"`` or ``"key=value"``
                - `id` (str): The id of the container.
                - `name` (str): The name of the container.
                - `ancestor` (str): Filter by container ancestor. Format of
                    ``<image-name>[:tag]``, ``<image-id>``, or
                    ``<image@digest>``.
                - `before` (str): Only containers created before a particular
                    container. Give the container name or id.
                - `since` (str): Only containers created after a particular
                    container. Give container name or id.

                A comprehensive list can be found in the documentation for
                `docker ps
                <https://docs.docker.com/engine/reference/commandline/ps>`_.

        Returns:
            (list of :py:class:`Container`)

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        resp = self.client.api.containers(all=all, before=before,
                                          filters=filters, limit=limit,
                                          since=since)
        return [self.get(r['Id']) for r in resp]

    def prune(self, filters=None):
        return self.client.api.prune_containers(filters=filters)
    prune.__doc__ = APIClient.prune_containers.__doc__


# kwargs to copy straight from run to create
RUN_CREATE_KWARGS = [
    'command',
    'detach',
    'domainname',
    'entrypoint',
    'environment',
    'healthcheck',
    'hostname',
    'image',
    'labels',
    'mac_address',
    'name',
    'network_disabled',
    'stdin_open',
    'stop_signal',
    'tty',
    'user',
    'volume_driver',
    'working_dir',
]

# kwargs to copy straight from run to host_config
RUN_HOST_CONFIG_KWARGS = [
    'blkio_weight_device',
    'blkio_weight',
    'cap_add',
    'cap_drop',
    'cgroup_parent',
    'cpu_count',
    'cpu_percent',
    'cpu_period',
    'cpu_quota',
    'cpu_shares',
    'cpuset_cpus',
    'cpuset_mems',
    'device_read_bps',
    'device_read_iops',
    'device_write_bps',
    'device_write_iops',
    'devices',
    'dns_opt',
    'dns_search',
    'dns',
    'extra_hosts',
    'group_add',
    'init',
    'init_path',
    'ipc_mode',
    'isolation',
    'kernel_memory',
    'links',
    'log_config',
    'lxc_conf',
    'mem_limit',
    'mem_reservation',
    'mem_swappiness',
    'memswap_limit',
    'nano_cpus',
    'network_mode',
    'oom_kill_disable',
    'oom_score_adj',
    'pid_mode',
    'pids_limit',
    'privileged',
    'publish_all_ports',
    'read_only',
    'restart_policy',
    'security_opt',
    'shm_size',
    'storage_opt',
    'sysctls',
    'tmpfs',
    'ulimits',
    'userns_mode',
    'version',
    'volumes_from',
]


def _create_container_args(kwargs):
    """
    Convert arguments to create() to arguments to create_container().
    """
    # Copy over kwargs which can be copied directly
    create_kwargs = {}
    for key in copy.copy(kwargs):
        if key in RUN_CREATE_KWARGS:
            create_kwargs[key] = kwargs.pop(key)
    host_config_kwargs = {}
    for key in copy.copy(kwargs):
        if key in RUN_HOST_CONFIG_KWARGS:
            host_config_kwargs[key] = kwargs.pop(key)

    # Process kwargs which are split over both create and host_config
    ports = kwargs.pop('ports', {})
    if ports:
        host_config_kwargs['port_bindings'] = ports

    volumes = kwargs.pop('volumes', {})
    if volumes:
        host_config_kwargs['binds'] = volumes

    network = kwargs.pop('network', None)
    if network:
        create_kwargs['networking_config'] = {network: None}
        host_config_kwargs['network_mode'] = network

    # All kwargs should have been consumed by this point, so raise
    # error if any are left
    if kwargs:
        raise create_unexpected_kwargs_error('run', kwargs)

    create_kwargs['host_config'] = HostConfig(**host_config_kwargs)

    # Fill in any kwargs which need processing by create_host_config first
    port_bindings = create_kwargs['host_config'].get('PortBindings')
    if port_bindings:
        # sort to make consistent for tests
        create_kwargs['ports'] = [tuple(p.split('/', 1))
                                  for p in sorted(port_bindings.keys())]
    binds = create_kwargs['host_config'].get('Binds')
    if binds:
        create_kwargs['volumes'] = [_host_volume_from_bind(v) for v in binds]
    return create_kwargs


def _host_volume_from_bind(bind):
    bits = bind.split(':')
    if len(bits) == 1:
        return bits[0]
    elif len(bits) == 2 and bits[1] in ('ro', 'rw'):
        return bits[0]
    else:
        return bits[1]




def add_port_mapping(port_bindings, internal_port, external):
    if internal_port in port_bindings:
        port_bindings[internal_port].append(external)
    else:
        port_bindings[internal_port] = [external]


def add_port(port_bindings, internal_port_range, external_range):
    if external_range is None:
        for internal_port in internal_port_range:
            add_port_mapping(port_bindings, internal_port, None)
    else:
        ports = zip(internal_port_range, external_range)
        for internal_port, external_port in ports:
            add_port_mapping(port_bindings, internal_port, external_port)


def build_port_bindings(ports):
    port_bindings = {}
    for port in ports:
        internal_port_range, external_range = split_port(port)
        add_port(port_bindings, internal_port_range, external_range)
    return port_bindings


def to_port_range(port, randomly_available_port=False):
    if not port:
        return None

    protocol = ""
    if "/" in port:
        parts = port.split("/")
        if len(parts) != 2:
            _raise_invalid_port(port)

        port, protocol = parts
        protocol = "/" + protocol

    if randomly_available_port:
        return ["%s%s" % (port, protocol)]

    parts = str(port).split('-')

    if len(parts) == 1:
        return ["%s%s" % (port, protocol)]

    if len(parts) == 2:
        full_port_range = range(int(parts[0]), int(parts[1]) + 1)
        return ["%s%s" % (p, protocol) for p in full_port_range]

    raise ValueError('Invalid port range "%s", should be '
                     'port or startport-endport' % port)


def _raise_invalid_port(port):
    raise ValueError('Invalid port "%s", should be '
                     '[[remote_ip:]remote_port[-remote_port]:]'
                     'port[/protocol]' % port)


def split_port(port):
    parts = str(port).split(':')

    if not 1 <= len(parts) <= 3:
        _raise_invalid_port(port)

    if len(parts) == 1:
        internal_port, = parts
        if not internal_port:
            _raise_invalid_port(port)
        return to_port_range(internal_port), None
    if len(parts) == 2:
        external_port, internal_port = parts

        internal_range = to_port_range(internal_port)
        if internal_range is None:
            _raise_invalid_port(port)

        external_range = to_port_range(external_port, len(internal_range) == 1)
        if external_range is None:
            _raise_invalid_port(port)

        if len(internal_range) != len(external_range):
            raise ValueError('Port ranges don\'t match in length')

        return internal_range, external_range

    external_ip, external_port, internal_port = parts

    if not internal_port:
        _raise_invalid_port(port)

    internal_range = to_port_range(internal_port)
    external_range = to_port_range(external_port, len(internal_range) == 1)

    if not external_range:
        external_range = [None] * len(internal_range)

    if len(internal_range) != len(external_range):
        raise ValueError('Port ranges don\'t match in length')

    return internal_range, [(external_ip, ex_port or None)
                            for ex_port in external_range]



import json
import socket

from functools import partial
from contextlib import contextmanager
from collections import deque

from ._requires.docker import APIClient, errors
from ._requires.docker.utils import kwargs_from_env


APIError = errors.APIError
NotFound = errors.NotFound


class ChunkedReader:

    def __init__(self, sock, *, loop):
        self.sock = sock
        self.loop = loop
        self.chunk_size = None
        self.tail = b''
        self.chunks = deque()
        self.complete = False

    async def read(self):
        while not self.chunks:
            if self.complete:
                raise RuntimeError('Stream is already consumed')
            data = await self.loop.sock_recv(self.sock, 4096)
            if not data:
                raise IOError('Incomplete response')
            self.tail += data
            while True:
                if self.chunk_size is None:
                    parts = self.tail.split(b'\r\n', 1)
                    if len(parts) == 2:
                        chunk_size_hex, self.tail = parts
                        self.chunk_size = int(chunk_size_hex, 16)
                    else:
                        break
                else:
                    if len(self.tail) >= self.chunk_size + 2:
                        self.chunks.append(self.tail[:self.chunk_size])
                        self.tail = self.tail[self.chunk_size + 2:]
                        if self.chunk_size == 0:
                            self.complete = True
                        self.chunk_size = None
                        continue
        return self.chunks.popleft()


class DockerStreamDecoder:

    def __init__(self, reader):
        self.reader = reader

    async def read(self):
        chunk = await self.reader.read()
        if chunk:
            return map(json.loads, chunk.decode('utf-8').strip().split('\r\n'))
        else:
            return []


class _Client(APIClient):

    def __init__(self, *args, loop, **kwargs):
        super().__init__(*args, **kwargs)
        self.loop = loop

    @contextmanager
    def _stream_ctx(self, response, decode=False):
        with self._get_raw_response_socket(response) as sock_io:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM,
                                 fileno=sock_io.fileno())
            sock.setblocking(False)
            reader = ChunkedReader(sock, loop=self.loop)
            if decode:
                reader = DockerStreamDecoder(reader)
            yield reader

    def _stream_helper(self, response, decode=False):
        if response.raw._fp.chunked:
            return self._stream_ctx(response, decode=decode)
        else:
            raise APIError('Error', response)


class _AsyncContextManagerAdapter:

    def __init__(self, func):
        self._func = func

    async def __aenter__(self):
        self._ctx = await self._func()
        return self._ctx.__enter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._ctx.__exit__(exc_type, exc_val, exc_tb)


class AsyncClient:

    def __init__(self, *, loop):
        self._client = _Client(version='auto', loop=loop, **kwargs_from_env())
        self._loop = loop

    def create_host_config(self, *args, **kwargs):
        return self._client.create_host_config(*args, **kwargs)

    def create_networking_config(self, *args, **kwargs):
        return self._client.create_networking_config(*args, **kwargs)

    async def _exec(self, func, *args, **kwargs):
        wrapper = partial(func, *args, **kwargs)
        result = await self._loop.run_in_executor(None, wrapper)
        return result

    def images(self, *args, **kwargs):
        return self._exec(self._client.images, *args, **kwargs)

    def build(self, *args, stream=False, **kwargs):
        def proc():
            return self._exec(self._client.build, *args,
                              stream=stream, **kwargs)
        return _AsyncContextManagerAdapter(proc) if stream else proc()

    def create_container(self, *args, **kwargs):
        return self._exec(self._client.create_container, *args, **kwargs)

    def start(self, *args, **kwargs):
        return self._exec(self._client.start, *args, **kwargs)

    def remove_container(self, *args, **kwargs):
        return self._exec(self._client.remove_container, *args, **kwargs)

    def put_archive(self, *args, **kwargs):
        return self._exec(self._client.put_archive, *args, **kwargs)

    def exec_create(self, *args, **kwargs):
        return self._exec(self._client.exec_create, *args, **kwargs)

    def exec_start(self, *args, **kwargs):
        return self._exec(self._client.exec_start, *args, **kwargs)

    def exec_inspect(self, *args, **kwargs):
        return self._exec(self._client.exec_inspect, *args, **kwargs)

    def pause(self, *args, **kwargs):
        return self._exec(self._client.pause, *args, **kwargs)

    def commit(self, *args, **kwargs):
        return self._exec(self._client.commit, *args, **kwargs)

    def unpause(self, *args, **kwargs):
        return self._exec(self._client.unpause, *args, **kwargs)

    def pull(self, *args, **kwargs):
        return self._exec(self._client.pull, *args, **kwargs)

    def push(self, *args, **kwargs):
        return self._exec(self._client.push, *args, **kwargs)

    def stop(self, *args, **kwargs):
        return self._exec(self._client.stop, *args, **kwargs)

    def kill(self, *args, **kwargs):
        return self._exec(self._client.kill, *args, **kwargs)

    def resize(self, *args, **kwargs):
        return self._exec(self._client.resize, *args, **kwargs)

    def attach_socket(self, *args, **kwargs):
        def proc():
            return self._exec(self._client.attach_socket, *args, **kwargs)
        return _AsyncContextManagerAdapter(proc)

    def wait(self, *args, **kwargs):
        return self._exec(self._client.wait, *args, **kwargs)

    def containers(self, *args, **kwargs):
        return self._exec(self._client.containers, *args, **kwargs)

    def create_network(self, *args, **kwargs):
        return self._exec(self._client.create_network, *args, **kwargs)

    def remove_image(self, *args, **kwargs):
        return self._exec(self._client.remove_image, *args, **kwargs)



"""
websocket - WebSocket client library for Python

Copyright (C) 2010 Hiroki Ohtani(liris)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA  02110-1335  USA

"""
import errno
import os
import socket
import sys

import pi._requires.six

from ._exceptions import *
from ._logging import *
from ._socket import*
from ._ssl_compat import *
from ._url import *

if pi._requires.six.PY3:
    from base64 import encodebytes as base64encode
else:
    from base64 import encodestring as base64encode

__all__ = ["proxy_info", "connect", "read_headers"]


class proxy_info(object):

    def __init__(self, **options):
        self.host = options.get("http_proxy_host", None)
        if self.host:
            self.port = options.get("http_proxy_port", 0)
            self.auth = options.get("http_proxy_auth", None)
            self.no_proxy = options.get("http_no_proxy", None)
        else:
            self.port = 0
            self.auth = None
            self.no_proxy = None


def connect(url, options, proxy, socket):
    hostname, port, resource, is_secure = parse_url(url)

    if socket:
        return socket, (hostname, port, resource)

    addrinfo_list, need_tunnel, auth = _get_addrinfo_list(
        hostname, port, is_secure, proxy)
    if not addrinfo_list:
        raise WebSocketException(
            "Host not found.: " + hostname + ":" + str(port))

    sock = None
    try:
        sock = _open_socket(addrinfo_list, options.sockopt, options.timeout)
        if need_tunnel:
            sock = _tunnel(sock, hostname, port, auth)

        if is_secure:
            if HAVE_SSL:
                sock = _ssl_socket(sock, options.sslopt, hostname)
            else:
                raise WebSocketException("SSL not available.")

        return sock, (hostname, port, resource)
    except:
        if sock:
            sock.close()
        raise


def _get_addrinfo_list(hostname, port, is_secure, proxy):
    phost, pport, pauth = get_proxy_info(
        hostname, is_secure, proxy.host, proxy.port, proxy.auth, proxy.no_proxy)
    if not phost:
        addrinfo_list = socket.getaddrinfo(
            hostname, port, 0, 0, socket.SOL_TCP)
        return addrinfo_list, False, None
    else:
        pport = pport and pport or 80
        addrinfo_list = socket.getaddrinfo(phost, pport, 0, 0, socket.SOL_TCP)
        return addrinfo_list, True, pauth


def _open_socket(addrinfo_list, sockopt, timeout):
    err = None
    for addrinfo in addrinfo_list:
        family = addrinfo[0]
        sock = socket.socket(family)
        sock.settimeout(timeout)
        for opts in DEFAULT_SOCKET_OPTION:
            sock.setsockopt(*opts)
        for opts in sockopt:
            sock.setsockopt(*opts)

        address = addrinfo[4]
        try:
            sock.connect(address)
        except socket.error as error:
            error.remote_ip = str(address[0])
            if error.errno in (errno.ECONNREFUSED, ):
                err = error
                continue
            else:
                raise
        else:
            break
    else:
        raise err

    return sock


def _can_use_sni():
    return pi._requires.six.PY2 and sys.version_info >= (2, 7, 9) or sys.version_info >= (3, 2)


def _wrap_sni_socket(sock, sslopt, hostname, check_hostname):
    context = ssl.SSLContext(sslopt.get('ssl_version', ssl.PROTOCOL_SSLv23))

    if sslopt.get('cert_reqs', ssl.CERT_NONE) != ssl.CERT_NONE:
        context.load_verify_locations(cafile=sslopt.get('ca_certs', None))
    if sslopt.get('certfile', None):
        context.load_cert_chain(
            sslopt['certfile'],
            sslopt.get('keyfile', None),
            sslopt.get('password', None),
        )
    # see
    # https://github.com/liris/websocket-client/commit/b96a2e8fa765753e82eea531adb19716b52ca3ca#commitcomment-10803153
    context.verify_mode = sslopt['cert_reqs']
    if HAVE_CONTEXT_CHECK_HOSTNAME:
        context.check_hostname = check_hostname
    if 'ciphers' in sslopt:
        context.set_ciphers(sslopt['ciphers'])
    if 'cert_chain' in sslopt:
        certfile, keyfile, password = sslopt['cert_chain']
        context.load_cert_chain(certfile, keyfile, password)

    return context.wrap_socket(
        sock,
        do_handshake_on_connect=sslopt.get('do_handshake_on_connect', True),
        suppress_ragged_eofs=sslopt.get('suppress_ragged_eofs', True),
        server_hostname=hostname,
    )


def _ssl_socket(sock, user_sslopt, hostname):
    sslopt = dict(cert_reqs=ssl.CERT_REQUIRED)
    sslopt.update(user_sslopt)

    if os.environ.get('WEBSOCKET_CLIENT_CA_BUNDLE'):
        certPath = os.environ.get('WEBSOCKET_CLIENT_CA_BUNDLE')
    else:
        certPath = os.path.join(
            os.path.dirname(__file__), "cacert.pem")
    if os.path.isfile(certPath) and user_sslopt.get('ca_certs', None) is None:
        sslopt['ca_certs'] = certPath
    check_hostname = sslopt["cert_reqs"] != ssl.CERT_NONE and sslopt.pop(
        'check_hostname', True)

    if _can_use_sni():
        sock = _wrap_sni_socket(sock, sslopt, hostname, check_hostname)
    else:
        sslopt.pop('check_hostname', True)
        sock = ssl.wrap_socket(sock, **sslopt)

    if not HAVE_CONTEXT_CHECK_HOSTNAME and check_hostname:
        match_hostname(sock.getpeercert(), hostname)

    return sock


def _tunnel(sock, host, port, auth):
    debug("Connecting proxy...")
    connect_header = "CONNECT %s:%d HTTP/1.0\r\n" % (host, port)
    # TODO: support digest auth.
    if auth and auth[0]:
        auth_str = auth[0]
        if auth[1]:
            auth_str += ":" + auth[1]
        encoded_str = base64encode(auth_str.encode()).strip().decode()
        connect_header += "Proxy-Authorization: Basic %s\r\n" % encoded_str
    connect_header += "\r\n"
    dump("request header", connect_header)

    send(sock, connect_header)

    try:
        status, resp_headers = read_headers(sock)
    except Exception as e:
        raise WebSocketProxyException(str(e))

    if status != 200:
        raise WebSocketProxyException(
            "failed CONNECT via proxy status: %r" % status)

    return sock


def read_headers(sock):
    status = None
    headers = {}
    trace("--- response header ---")

    while True:
        line = recv_line(sock)
        line = line.decode('utf-8').strip()
        if not line:
            break
        trace(line)
        if not status:

            status_info = line.split(" ", 2)
            status = int(status_info[1])
        else:
            kv = line.split(":", 1)
            if len(kv) == 2:
                key, value = kv
                headers[key.lower()] = value.strip()
            else:
                raise WebSocketException("Invalid header")

    trace("-----------------------")

    return status, headers



import logging
from pi._requires.six.moves import http_client
from .. import types
from .. import utils
log = logging.getLogger(__name__)


class SwarmApiMixin(object):

    def create_swarm_spec(self, *args, **kwargs):
        """
        Create a ``docker.types.SwarmSpec`` instance that can be used as the
        ``swarm_spec`` argument in
        :py:meth:`~docker.api.swarm.SwarmApiMixin.init_swarm`.

        Args:
            task_history_retention_limit (int): Maximum number of tasks
                history stored.
            snapshot_interval (int): Number of logs entries between snapshot.
            keep_old_snapshots (int): Number of snapshots to keep beyond the
                current snapshot.
            log_entries_for_slow_followers (int): Number of log entries to
                keep around to sync up slow followers after a snapshot is
                created.
            heartbeat_tick (int): Amount of ticks (in seconds) between each
                heartbeat.
            election_tick (int): Amount of ticks (in seconds) needed without a
                leader to trigger a new election.
            dispatcher_heartbeat_period (int):  The delay for an agent to send
                a heartbeat to the dispatcher.
            node_cert_expiry (int): Automatic expiry for nodes certificates.
            external_ca (dict): Configuration for forwarding signing requests
                to an external certificate authority. Use
                ``docker.types.SwarmExternalCA``.
            name (string): Swarm's name

        Returns:
            ``docker.types.SwarmSpec`` instance.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.

        Example:

            >>> spec = client.create_swarm_spec(
              snapshot_interval=5000, log_entries_for_slow_followers=1200
            )
            >>> client.init_swarm(
              advertise_addr='eth0', listen_addr='0.0.0.0:5000',
              force_new_cluster=False, swarm_spec=spec
            )
        """
        return types.SwarmSpec(*args, **kwargs)

    @utils.minimum_version('1.24')
    def init_swarm(self, advertise_addr=None, listen_addr='0.0.0.0:2377',
                   force_new_cluster=False, swarm_spec=None):
        """
        Initialize a new Swarm using the current connected engine as the first
        node.

        Args:
            advertise_addr (string): Externally reachable address advertised
                to other nodes. This can either be an address/port combination
                in the form ``192.168.1.1:4567``, or an interface followed by a
                port number, like ``eth0:4567``. If the port number is omitted,
                the port number from the listen address is used. If
                ``advertise_addr`` is not specified, it will be automatically
                detected when possible. Default: None
            listen_addr (string): Listen address used for inter-manager
                communication, as well as determining the networking interface
                used for the VXLAN Tunnel Endpoint (VTEP). This can either be
                an address/port combination in the form ``192.168.1.1:4567``,
                or an interface followed by a port number, like ``eth0:4567``.
                If the port number is omitted, the default swarm listening port
                is used. Default: '0.0.0.0:2377'
            force_new_cluster (bool): Force creating a new Swarm, even if
                already part of one. Default: False
            swarm_spec (dict): Configuration settings of the new Swarm. Use
                ``APIClient.create_swarm_spec`` to generate a valid
                configuration. Default: None

        Returns:
            ``True`` if successful.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """

        url = self._url('/swarm/init')
        if swarm_spec is not None and not isinstance(swarm_spec, dict):
            raise TypeError('swarm_spec must be a dictionary')
        data = {
            'AdvertiseAddr': advertise_addr,
            'ListenAddr': listen_addr,
            'ForceNewCluster': force_new_cluster,
            'Spec': swarm_spec,
        }
        response = self._post_json(url, data=data)
        self._raise_for_status(response)
        return True

    @utils.minimum_version('1.24')
    def inspect_swarm(self):
        """
        Retrieve low-level information about the current swarm.

        Returns:
            A dictionary containing data about the swarm.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        url = self._url('/swarm')
        return self._result(self._get(url), True)

    @utils.check_resource
    @utils.minimum_version('1.24')
    def inspect_node(self, node_id):
        """
        Retrieve low-level information about a swarm node

        Args:
            node_id (string): ID of the node to be inspected.

        Returns:
            A dictionary containing data about this node.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        url = self._url('/nodes/{0}', node_id)
        return self._result(self._get(url), True)

    @utils.minimum_version('1.24')
    def join_swarm(self, remote_addrs, join_token, listen_addr=None,
                   advertise_addr=None):
        """
        Make this Engine join a swarm that has already been created.

        Args:
            remote_addrs (:py:class:`list`): Addresses of one or more manager
                nodes already participating in the Swarm to join.
            join_token (string): Secret token for joining this Swarm.
            listen_addr (string): Listen address used for inter-manager
                communication if the node gets promoted to manager, as well as
                determining the networking interface used for the VXLAN Tunnel
                Endpoint (VTEP). Default: ``None``
            advertise_addr (string): Externally reachable address advertised
                to other nodes. This can either be an address/port combination
                in the form ``192.168.1.1:4567``, or an interface followed by a
                port number, like ``eth0:4567``. If the port number is omitted,
                the port number from the listen address is used. If
                AdvertiseAddr is not specified, it will be automatically
                detected when possible. Default: ``None``

        Returns:
            ``True`` if the request went through.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        data = {
            "RemoteAddrs": remote_addrs,
            "ListenAddr": listen_addr,
            "JoinToken": join_token,
            "AdvertiseAddr": advertise_addr,
        }
        url = self._url('/swarm/join')
        response = self._post_json(url, data=data)
        self._raise_for_status(response)
        return True

    @utils.minimum_version('1.24')
    def leave_swarm(self, force=False):
        """
        Leave a swarm.

        Args:
            force (bool): Leave the swarm even if this node is a manager.
                Default: ``False``

        Returns:
            ``True`` if the request went through.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        url = self._url('/swarm/leave')
        response = self._post(url, params={'force': force})
        # Ignore "this node is not part of a swarm" error
        if force and response.status_code == http_client.NOT_ACCEPTABLE:
            return True
        # FIXME: Temporary workaround for 1.13.0-rc bug
        # https://github.com/docker/docker/issues/29192
        if force and response.status_code == http_client.SERVICE_UNAVAILABLE:
            return True
        self._raise_for_status(response)
        return True

    @utils.minimum_version('1.24')
    def nodes(self, filters=None):
        """
        List swarm nodes.

        Args:
            filters (dict): Filters to process on the nodes list. Valid
                filters: ``id``, ``name``, ``membership`` and ``role``.
                Default: ``None``

        Returns:
            A list of dictionaries containing data about each swarm node.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        url = self._url('/nodes')
        params = {}
        if filters:
            params['filters'] = utils.convert_filters(filters)

        return self._result(self._get(url, params=params), True)

    @utils.check_resource
    @utils.minimum_version('1.24')
    def remove_node(self, node_id, force=False):
        """
        Remove a node from the swarm.

        Args:
            node_id (string): ID of the node to be removed.
            force (bool): Force remove an active node. Default: `False`

        Raises:
            :py:class:`docker.errors.NotFound`
                If the node referenced doesn't exist in the swarm.

            :py:class:`docker.errors.APIError`
                If the server returns an error.
        Returns:
            `True` if the request was successful.
        """
        url = self._url('/nodes/{0}', node_id)
        params = {
            'force': force
        }
        res = self._delete(url, params=params)
        self._raise_for_status(res)
        return True

    @utils.minimum_version('1.24')
    def update_node(self, node_id, version, node_spec=None):
        """
        Update the Node's configuration

        Args:

            node_id (string): ID of the node to be updated.
            version (int): The version number of the node object being
                updated. This is required to avoid conflicting writes.
            node_spec (dict): Configuration settings to update. Any values
                not provided will be removed. Default: ``None``

        Returns:
            `True` if the request went through.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.

        Example:

            >>> node_spec = {'Availability': 'active',
                         'Name': 'node-name',
                         'Role': 'manager',
                         'Labels': {'foo': 'bar'}
                        }
            >>> client.update_node(node_id='24ifsmvkjbyhk', version=8,
                node_spec=node_spec)

        """
        url = self._url('/nodes/{0}/update?version={1}', node_id, str(version))
        res = self._post_json(url, data=node_spec)
        self._raise_for_status(res)
        return True

    @utils.minimum_version('1.24')
    def update_swarm(self, version, swarm_spec=None, rotate_worker_token=False,
                     rotate_manager_token=False):
        """
        Update the Swarm's configuration

        Args:
            version (int): The version number of the swarm object being
                updated. This is required to avoid conflicting writes.
            swarm_spec (dict): Configuration settings to update. Use
                :py:meth:`~docker.api.swarm.SwarmApiMixin.create_swarm_spec` to
                generate a valid configuration. Default: ``None``.
            rotate_worker_token (bool): Rotate the worker join token. Default:
                ``False``.
            rotate_manager_token (bool): Rotate the manager join token.
                Default: ``False``.

        Returns:
            ``True`` if the request went through.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """

        url = self._url('/swarm/update')
        response = self._post_json(url, data=swarm_spec, params={
            'rotateWorkerToken': rotate_worker_token,
            'rotateManagerToken': rotate_manager_token,
            'version': version
        })
        self._raise_for_status(response)
        return True



import functools

from .. import errors
from . import utils


def check_resource(f):
    @functools.wraps(f)
    def wrapped(self, resource_id=None, *args, **kwargs):
        if resource_id is None:
            if kwargs.get('container'):
                resource_id = kwargs.pop('container')
            elif kwargs.get('image'):
                resource_id = kwargs.pop('image')
        if isinstance(resource_id, dict):
            resource_id = resource_id.get('Id', resource_id.get('ID'))
        if not resource_id:
            raise errors.NullResource(
                'Resource ID was not provided'
            )
        return f(self, resource_id, *args, **kwargs)
    return wrapped


def minimum_version(version):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(self, *args, **kwargs):
            if utils.version_lt(self._version, version):
                raise errors.InvalidVersion(
                    '{0} is not available for version < {1}'.format(
                        f.__name__, version
                    )
                )
            return f(self, *args, **kwargs)
        return wrapper
    return decorator


def update_headers(f):
    def inner(self, *args, **kwargs):
        if 'HttpHeaders' in self._auth_configs:
            if not kwargs.get('headers'):
                kwargs['headers'] = self._auth_configs['HttpHeaders']
            else:
                kwargs['headers'].update(self._auth_configs['HttpHeaders'])
        return f(self, *args, **kwargs)
    return inner



from __future__ import absolute_import, division, print_function

from ._funcs import (
    asdict,
    assoc,
    astuple,
    has,
)
from ._make import (
    Attribute,
    Factory,
    NOTHING,
    attr,
    attributes,
    fields,
    make_class,
    validate,
)
from ._config import (
    get_run_validators,
    set_run_validators,
)
from . import exceptions
from . import filters
from . import validators


__version__ = "16.2.0"

__title__ = "attrs"
__description__ = "Attributes Without Boilerplate"
__uri__ = "https://attrs.readthedocs.io/"
__doc__ = __description__ + " <" + __uri__ + ">"

__author__ = "Hynek Schlawack"
__email__ = "hs@ox.cx"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2015 Hynek Schlawack"


s = attrs = attributes
ib = attrib = attr

__all__ = [
    "Attribute",
    "Factory",
    "NOTHING",
    "asdict",
    "astuple",
    "assoc",
    "attr",
    "attrib",
    "attributes",
    "attrs",
    "exceptions",
    "fields",
    "filters",
    "get_run_validators",
    "has",
    "ib",
    "make_class",
    "s",
    "set_run_validators",
    "validate",
    "validators",
]



import os

from ..constants import IS_WINDOWS_PLATFORM
from .fnmatch import fnmatch
from .utils import create_archive


def tar(path, exclude=None, dockerfile=None, fileobj=None, gzip=False):
    root = os.path.abspath(path)
    exclude = exclude or []

    return create_archive(
        files=sorted(exclude_paths(root, exclude, dockerfile=dockerfile)),
        root=root, fileobj=fileobj, gzip=gzip
    )


def exclude_paths(root, patterns, dockerfile=None):
    """
    Given a root directory path and a list of .dockerignore patterns, return
    an iterator of all paths (both regular files and directories) in the root
    directory that do *not* match any of the patterns.

    All paths returned are relative to the root.
    """
    if dockerfile is None:
        dockerfile = 'Dockerfile'

    exceptions = [p for p in patterns if p.startswith('!')]

    include_patterns = [p[1:] for p in exceptions]
    include_patterns += [dockerfile, '.dockerignore']

    exclude_patterns = list(set(patterns) - set(exceptions))

    paths = get_paths(root, exclude_patterns, include_patterns,
                      has_exceptions=len(exceptions) > 0)

    return set(paths).union(
        # If the Dockerfile is in a subdirectory that is excluded, get_paths
        # will not descend into it and the file will be skipped. This ensures
        # it doesn't happen.
        set([dockerfile.replace('/', os.path.sep)])
        if os.path.exists(os.path.join(root, dockerfile)) else set()
    )


def should_include(path, exclude_patterns, include_patterns):
    """
    Given a path, a list of exclude patterns, and a list of inclusion patterns:

    1. Returns True if the path doesn't match any exclusion pattern
    2. Returns False if the path matches an exclusion pattern and doesn't match
       an inclusion pattern
    3. Returns true if the path matches an exclusion pattern and matches an
       inclusion pattern
    """
    for pattern in exclude_patterns:
        if match_path(path, pattern):
            for pattern in include_patterns:
                if match_path(path, pattern):
                    return True
            return False
    return True


def should_check_directory(directory_path, exclude_patterns, include_patterns):
    """
    Given a directory path, a list of exclude patterns, and a list of inclusion
    patterns:

    1. Returns True if the directory path should be included according to
       should_include.
    2. Returns True if the directory path is the prefix for an inclusion
       pattern
    3. Returns False otherwise
    """

    # To account for exception rules, check directories if their path is a
    # a prefix to an inclusion pattern. This logic conforms with the current
    # docker logic (2016-10-27):
    # https://github.com/docker/docker/blob/bc52939b0455116ab8e0da67869ec81c1a1c3e2c/pkg/archive/archive.go#L640-L671

    def normalize_path(path):
        return path.replace(os.path.sep, '/')

    path_with_slash = normalize_path(directory_path) + '/'
    possible_child_patterns = [
        pattern for pattern in map(normalize_path, include_patterns)
        if (pattern + '/').startswith(path_with_slash)
    ]
    directory_included = should_include(
        directory_path, exclude_patterns, include_patterns
    )
    return directory_included or len(possible_child_patterns) > 0


def get_paths(root, exclude_patterns, include_patterns, has_exceptions=False):
    paths = []

    for parent, dirs, files in os.walk(root, topdown=True, followlinks=False):
        parent = os.path.relpath(parent, root)
        if parent == '.':
            parent = ''

        # Remove excluded patterns from the list of directories to traverse
        # by mutating the dirs we're iterating over.
        # This looks strange, but is considered the correct way to skip
        # traversal. See https://docs.python.org/2/library/os.html#os.walk
        dirs[:] = [
            d for d in dirs if should_check_directory(
                os.path.join(parent, d), exclude_patterns, include_patterns
            )
        ]

        for path in dirs:
            if should_include(os.path.join(parent, path),
                              exclude_patterns, include_patterns):
                paths.append(os.path.join(parent, path))

        for path in files:
            if should_include(os.path.join(parent, path),
                              exclude_patterns, include_patterns):
                paths.append(os.path.join(parent, path))

    return paths


def match_path(path, pattern):
    pattern = pattern.rstrip('/' + os.path.sep)
    if pattern:
        pattern = os.path.relpath(pattern)

    pattern_components = pattern.split(os.path.sep)
    if len(pattern_components) == 1 and IS_WINDOWS_PLATFORM:
        pattern_components = pattern.split('/')

    if '**' not in pattern:
        path_components = path.split(os.path.sep)[:len(pattern_components)]
    else:
        path_components = path.split(os.path.sep)
    return fnmatch('/'.join(path_components), '/'.join(pattern_components))



import os
import math
import uuid
import tarfile
import hashlib
import tempfile
import unicodedata

from pathlib import Path
from asyncio import wait, Queue, Event, gather, FIRST_EXCEPTION

from concurrent.futures import ProcessPoolExecutor

from ._requires import attr
from ._requires import jinja2
from ._requires import requests

from .types import ActionType
from .utils import terminate
from .images import docker_image, image_versions


class ResultBase:

    def value(self):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError


@attr.s
class Result(ResultBase):
    file = attr.ib()
    uuid = attr.ib(default=attr.Factory(lambda: uuid.uuid4().hex))

    def value(self):
        return hashlib.sha1(self.file.name.encode('utf-8')).hexdigest()

    def close(self):
        self.file.close()


@attr.s
class SimpleResult(ResultBase):
    content = attr.ib()

    def value(self):
        return self.content

    def close(self):
        pass


@attr.s
class ActionState:
    complete = attr.ib()
    result = attr.ib()
    error = attr.ib(default=None)


def download(url, file_name, destination):
    with tempfile.NamedTemporaryFile() as tmp:
        with tarfile.open(file_name, mode='w:tar') as tar:
            # downloaded = 0
            response = requests.get(url)
            response.raise_for_status()
            for chunk in response.iter_content(64 * 1024):
                tmp.write(chunk)
                # downloaded += len(chunk)
            tmp.seek(0)
            tar.addfile(tar.gettarinfo(arcname=destination, fileobj=tmp),
                        fileobj=tmp)


def file_(path, file_name, destination):
    file_path = Path(path).resolve()
    assert file_path.is_file(), file_path

    cur_path = Path('.').resolve()
    assert cur_path in file_path.parents, file_path

    with open(str(file_path), 'rb') as f:
        with tarfile.open(file_name, mode='w:tar') as tar:
            tar.addfile(tar.gettarinfo(arcname=destination, fileobj=f),
                        fileobj=f)


def bundle(action_path, file_name, destination):
    dir_path = Path(action_path).resolve()  # FIXME: raises FileNotFoundError
    assert dir_path.is_dir(), dir_path

    cur_path = Path('.').resolve()
    assert cur_path in dir_path.parents, dir_path

    def _arc_path(path_):
        return os.path.join(destination, unicodedata.normalize('NFC', path_))

    with tarfile.open(file_name, mode='w:tar') as tar:
        for abs_path, _, names in os.walk(str(dir_path)):
            rel_path = Path(abs_path).relative_to(dir_path)
            if rel_path != Path('.'):
                tar.addfile(tar.gettarinfo(abs_path, _arc_path(str(rel_path))))

            for name in names:
                file_abs_path = os.path.join(abs_path, name)
                file_rel_path = str(rel_path.joinpath(name))
                with open(file_abs_path, 'rb') as f:
                    tar.addfile(
                        tar.gettarinfo(arcname=_arc_path(str(file_rel_path)),
                                       fileobj=f),
                        fileobj=f,
                    )


def iter_actions(task):
    for value in task.where.values():
        if isinstance(value, ActionType):
            yield value


def get_action_states(tasks, *, loop):
    actions = {}
    for task in tasks:
        for action in iter_actions(task):
            if action in actions:
                continue
            result = Result(tempfile.NamedTemporaryFile())
            actions[action] = ActionState(Event(loop=loop), result)
    return actions


async def wait_actions(states, *, loop):
    if states:
        await gather(*[state.complete.wait() for state in states.values()],
                     loop=loop)


class ActionDispatcher:

    def __init__(self, states, io_queue, cpu_queue):
        self.states = states
        self.io_queue = io_queue
        self.cpu_queue = cpu_queue

    @classmethod
    async def dispatch(cls, states, io_queue, cpu_queue):
        dispatcher = cls(states, io_queue, cpu_queue)
        for action in states:
            await dispatcher.visit(action)

    def visit(self, obj):
        return obj.accept(self)

    async def visit_download(self, obj):
        await self.io_queue.put((obj, self.states[obj]))

    async def visit_file(self, obj):
        await self.cpu_queue.put((obj, self.states[obj]))

    async def visit_bundle(self, obj):
        await self.cpu_queue.put((obj, self.states[obj]))


class IOExecutor:

    def __init__(self, *, loop):
        self.loop = loop

    def visit(self, action):
        return action.accept(self)

    async def download(self, action, state):
        try:
            await self.loop.run_in_executor(
                None,
                download, action.url, state.result.file.name, state.result.uuid
            )
        except Exception as err:
            state.error = str(err)
            raise
        finally:
            state.complete.set()

    def visit_download(self, obj):
        return self.download


class CPUExecutor:

    def __init__(self, process_pool, *, loop):
        self.process_pool = process_pool
        self.loop = loop

    def visit(self, action):
        return action.accept(self)

    async def file(self, action, state):
        try:
            await self.loop.run_in_executor(
                self.process_pool,
                file_, action.path, state.result.file.name, state.result.uuid
            )
        except Exception as err:
            state.error = str(err)
        finally:
            state.complete.set()

    async def bundle(self, action, state):
        try:
            await self.loop.run_in_executor(
                self.process_pool,
                bundle, action.path, state.result.file.name, state.result.uuid
            )
        except Exception as err:
            state.error = str(err)
        finally:
            state.complete.set()

    def visit_file(self, obj):
        return self.file

    def visit_bundle(self, obj):
        return self.bundle


async def worker(queue, executor):
    while True:
        action, state = await queue.get()
        process = executor.visit(action)
        await process(action, state)


async def pool(queue, executor, concurrency=2, *, loop):
    pending = [loop.create_task(worker(queue, executor))
               for _ in range(concurrency)]
    try:
        # fast exit on first error
        _, pending = await wait(pending, loop=loop,
                                return_when=FIRST_EXCEPTION)
    finally:
        for task in pending:
            await terminate(task, loop=loop)


def task_cmd(task, results):
    ctx = {key: value if not isinstance(value, ActionType) else results[value]
           for key, value in task.where.items()}
    t = jinja2.Template(task.run)
    return t.render(ctx)


async def _exec(client, container, cmd):
    if isinstance(cmd, str):
        cmd = ['/bin/sh', '-c', cmd]
    exec_id = await client.exec_create(container, cmd)
    output = await client.exec_start(exec_id)
    info = await client.exec_inspect(exec_id)
    exit_code = info['ExitCode']
    if exit_code:
        print(output.decode('utf-8'))  # FIXME: proper output
    return exit_code


async def build(client, images_map, image, *, loop):
    version, = image_versions(images_map, [image])
    from_ = docker_image(images_map, image.from_)

    io_queue = Queue()
    io_executor = IOExecutor(loop=loop)

    process_pool = ProcessPoolExecutor()
    cpu_queue = Queue()
    cpu_executor = CPUExecutor(process_pool, loop=loop)

    states = get_action_states(image.tasks, loop=loop)
    submitted_states = set()

    io_pool_task = loop.create_task(pool(io_queue, io_executor, loop=loop))
    cpu_pool_task = loop.create_task(pool(cpu_queue, cpu_executor, loop=loop))

    c = await client.create_container(
        from_.name, '/bin/sh', detach=True, tty=True,
    )
    try:
        await client.start(c)
        exit_code = await _exec(client, c, ['mkdir', '/.pi'])
        if exit_code:
            return False

        await ActionDispatcher.dispatch(states, io_queue, cpu_queue)

        total = len(image.tasks)
        padding = math.ceil(math.log10(total + 1))

        for i, task in enumerate(image.tasks, 1):
            task_states = {action: states[action]
                           for action in iter_actions(task)}

            await wait_actions(task_states, loop=loop)

            errors = {action: state.error
                      for action, state in task_states.items()
                      if state.error is not None}
            if errors:
                raise Exception(repr(errors))

            task_results = {action: '/.pi/{}'.format(state.result.uuid)
                            for action, state in task_states.items()}

            for action, state in task_states.items():
                if action not in submitted_states:
                    with open(state.result.file.name, 'rb') as tar:
                        await client.put_archive(c, '/.pi', tar)
                    submitted_states.add(action)

            cmd = task_cmd(task, task_results)
            current_index = '{{:{}d}}'.format(padding).format(i)
            print('[{}/{}] {}'.format(current_index, total, cmd))
            exit_code = await _exec(client, c, cmd)
            if exit_code:
                return False

        exit_code = await _exec(client, c, ['rm', '-rf', '/.pi'])
        if exit_code:
            return False

        await client.pause(c)
        await client.commit(c, image.repository, version)
        await client.unpause(c)
        return True

    finally:
        await terminate(io_pool_task, loop=loop)
        await terminate(cpu_pool_task, loop=loop)
        process_pool.shutdown()
        for state in states.values():
            state.result.close()
        await client.remove_container(c, v=True, force=True)



import sys

from .._requires import click
from .._requires.tabulate import tabulate

from ..run import start
from ..utils import search_container, sh_to_list
from ..images import docker_image
from ..environ import async_cmd
from ..network import ensure_network
from ..console import pretty
from ..services import get_volumes, service_label

from .common import ExtGroup


@click.command('start', help='Start service')
@click.pass_obj
@async_cmd
async def service_start(env):
    name = click.get_current_context().meta['image_name']
    service = env.services.get(name, None)
    if service is None:
        click.echo('Unknown service name: {}'.format(name))
        sys.exit(-1)

    label = service_label(env.namespace, service)
    containers = await env.client.containers(all=True)
    container = next(search_container(label, containers), None)
    if container is not None:
        if container['State'] == 'running':
            click.echo('Service is already running')
        elif container['State'] == 'exited':
            await env.client.start(container)
            click.echo('Started previously stopped service')
        else:
            raise NotImplementedError(container['State'])
    else:
        exec_ = sh_to_list(service.exec) if service.exec else None
        args = sh_to_list(service.args) if service.args else None
        di = docker_image(env, service.image)
        await ensure_network(env.client, env.network)

        await start(env.client, di, args,
                    entrypoint=exec_,
                    volumes=get_volumes(service.volumes),
                    ports=service.ports,
                    environ=service.environ,
                    network=env.network,
                    network_alias=service.network_name or service.name,
                    label=label)
        click.echo('Service started')


@click.command('stop', help='Stop service')
@click.pass_obj
@async_cmd
async def service_stop(env):
    name = click.get_current_context().meta['image_name']
    service = env.services.get(name, None)
    if service is None:
        click.echo('Unknown service name: {}'.format(name))
        sys.exit(-1)

    label = service_label(env.namespace, service)
    all_containers = await env.client.containers(all=True)
    containers = list(search_container(label, all_containers))
    if not containers:
        click.echo('Service was not started')
        sys.exit(-1)

    for container in containers:
        if container['State'] == 'running':
            await env.client.stop(container, timeout=3)
        await env.client.remove_container(container, v=True, force=True)
    click.echo('Service stopped')


@async_cmd
async def _service_status(env):
    containers = await env.client.containers(all=True)

    running = set()
    exited = set()
    images = {}
    for container in containers:
        if container['State'] == 'exited':
            exited.update(container['Labels'])
        elif container['State'] == 'running':
            running.update(container['Labels'])
        for label in container['Labels']:
            images[label] = container['Image']

    rows = []
    for service in env.services:
        label = service_label(env.namespace, service)
        if label in running:
            status = pretty('{_green}running{_r}')
            image = images[label]
        elif label in exited:
            status = pretty('{_red}stopped{_r}')
            image = images[label]
        else:
            status = None
            image = None

        di = docker_image(env.images, service.image)

        if image is not None and image != di.name:
            image += ' (obsolete)'

        rows.append([service.name, status, image])
    click.echo(tabulate(rows, headers=['Service name', 'Status',
                                       'Docker image']))


def _service_ext_help(ctx, formatter):
    if ctx.obj.services:
        with formatter.section('Services'):
            formatter.write_dl([(service.name, service.description or '')
                                for service in ctx.obj.services])
    else:
        with formatter.section('Services'):
            formatter.write_text('--- not defined ---')


def _service_status_callback(ctx, param, value):
    if value and not ctx.resilient_parsing:
        _service_status(ctx.obj)
        ctx.exit()


def _service_callback(name):
    click.get_current_context().meta['image_name'] = name


def create_service_cli():
    params = [
        click.Option(['-s', '--status'], is_flag=True, is_eager=True,
                     expose_value=False, callback=_service_status_callback,
                     help='Display services status'),
        click.Argument(['name']),
    ]

    service_group = ExtGroup('service', params=params,
                             callback=_service_callback,
                             help='Services status and management',
                             ext_help=_service_ext_help)
    service_group.add_command(service_start)
    service_group.add_command(service_stop)

    cli = click.Group()
    cli.add_command(service_group)
    return cli



"""
    Configuration
    ~~~~~~~~~~~~~

    `Pi` uses tagged values feature of the YAML_ format, which are mapped
    directly into internal types of `Pi`. This kind of DSL is used to
    describe project environment in a most simple way.

    .. _YAML: http://yaml.org/spec/
"""
from enum import Enum
from typing import Optional, Union, Any, Sequence, Mapping  # noqa
from collections import OrderedDict

from ._requires import attr

from .utils import ImmutableDict


class EnumConstruct:

    @classmethod
    def construct(cls, loader, node):
        assert not loader.construct_scalar(node),\
            '{}: No arguments expected'.format(cls.__name__)
        return cls(node.tag)


class ScalarConstruct:

    @classmethod
    def construct(cls, loader, node):
        return cls(loader.construct_scalar(node))


class SequenceConstruct:

    @classmethod
    def construct(cls, loader, node):
        return cls(loader.construct_sequence(node, deep=True))


class MappingConstruct:
    __rename_to__ = ImmutableDict()

    @classmethod
    def construct(cls, loader, node):
        rename_from = {v: k for k, v in cls.__rename_to__.items()}

        params = loader.construct_mapping(node, deep=True)

        required = {rename_from.get(a.name, a.name)
                    for a in cls.__attrs_attrs__
                    if a.default is attr.NOTHING}

        optional = {rename_from.get(a.name, a.name)
                    for a in cls.__attrs_attrs__
                    if a.default is not attr.NOTHING}

        missing = set(required).difference(params)
        if missing:
            raise TypeError('Missing params {!r} for {!r}'.format(missing, cls))

        clean_params = {cls.__rename_to__.get(k, k): v
                        for k, v in params.items()
                        if k in required or k in optional}

        return cls(**clean_params)


@attr.s
class Meta(MappingConstruct):
    """``!Meta`` -- Project-specific settings

    :param str namespace: used to namespace entities to avoid names collision
        in different projects on the same host
    :param str description: description of the project, showed in the "help" of
        the project's CLI
    """
    __tag__ = '!Meta'

    namespace = attr.ib(default=None)  # type: Optional[str]
    description = attr.ib(default=None)  # type: Optional[str]

    def accept(self, visitor):
        return visitor.visit_meta(self)


@attr.s
class DockerImage(ScalarConstruct):
    __tag__ = '!DockerImage'

    name = attr.ib()

    def accept(self, visitor):
        return visitor.visit_dockerimage(self)

    @classmethod
    def from_image(cls, image, version):
        return cls('{}:{}'.format(image.repository, version))


@attr.s
class Task:
    run = attr.ib()
    where = attr.ib(default=attr.Factory(dict))

    def accept(self, visitor):
        return visitor.visit_task(self)

    @classmethod
    def from_config(cls, d):
        run = d['run']
        where = OrderedDict((k, v) for k, v in d.items()
                            if k != 'run')
        return cls(run, where)


def _convert_tasks(tasks):
    return [Task.from_config(d) for d in tasks]


@attr.s
class Image(MappingConstruct):
    __tag__ = '!Image'
    __rename_to__ = ImmutableDict([
        ('from', 'from_'),
    ])

    name = attr.ib()  # type: str
    repository = attr.ib()  # type: str
    from_ = attr.ib(default=None)  # type: Optional[Union[str, DockerImage]]
    # type: Sequence[Task]
    tasks = attr.ib(default=tuple(), hash=False, convert=_convert_tasks)
    description = attr.ib(default=None)  # type: Optional[str]

    def accept(self, visitor):
        return visitor.visit_image(self)


class VolumeType:

    def accept(self, visitor):
        raise NotImplementedError


class Mode(EnumConstruct, Enum):
    RO = '!RO'
    RW = '!RW'

    def accept(self, visitor):
        return getattr(visitor, 'visit_{}'.format(self.name))(self)


@attr.s
class LocalPath(VolumeType, MappingConstruct):
    __tag__ = '!LocalPath'
    __rename_to__ = ImmutableDict([
        ('from', 'from_'),
    ])

    from_ = attr.ib()  # type: str
    to = attr.ib()  # type: str
    mode = attr.ib(default=Mode.RO)  # type: Mode

    def accept(self, visitor):
        return visitor.visit_localpath(self)


@attr.s
class NamedVolume(VolumeType, MappingConstruct):
    __tag__ = '!NamedVolume'

    name = attr.ib()  # type: str
    to = attr.ib()  # type: str
    mode = attr.ib(default=Mode.RO)  # type: Mode

    def accept(self, visitor):
        return visitor.visit_namedvolume(self)


@attr.s
class Expose(MappingConstruct):
    __tag__ = '!Expose'
    __rename_to__ = ImmutableDict([
        ('as', 'as_'),
    ])

    port = attr.ib()  # type: int
    as_ = attr.ib()  # type: int
    addr = attr.ib(default='127.0.0.1')  # type: str
    proto = attr.ib(default='tcp')  # type: str

    def accept(self, visitor):
        return visitor.visit_expose(self)


@attr.s
class Service(MappingConstruct):
    __tag__ = '!Service'
    __rename_to__ = ImmutableDict([
        ('network-name', 'network_name'),
    ])

    name = attr.ib()  # type: str
    image = attr.ib()  # type: Union[str, DockerImage]
    volumes = attr.ib(default=None)  # type: Optional[Sequence[VolumeType]]
    ports = attr.ib(default=None)  # type: Optional[Sequence[Expose]]
    environ = attr.ib(default=None)  # type: Optional[Mapping[str: str]]
    requires = attr.ib(default=None)  # type: Optional[Sequence[str]]
    exec = attr.ib(default=None)  # type: Union[str, Sequence[str]]
    args = attr.ib(default=None)  # type: Union[str, Sequence[str]]
    network_name = attr.ib(default=None)  # type: Optional[str]
    description = attr.ib(default=None)  # type: Optional[str]

    def accept(self, visitor):
        return visitor.visit_service(self)


@attr.s
class ParameterType:
    name = attr.ib()  # type: str
    type = attr.ib(default=None)  # type: Optional[str]
    default = attr.ib(default=None)  # type: Any

    def accept(self, visitor):
        raise NotImplementedError


class Argument(ParameterType, MappingConstruct):
    __tag__ = '!Argument'

    def accept(self, visitor):
        return visitor.visit_argument(self)


class Option(ParameterType, MappingConstruct):
    __tag__ = '!Option'

    def accept(self, visitor):
        return visitor.visit_option(self)


class CommandType:

    def accept(self, visitor):
        raise NotImplementedError


@attr.s
class Command(CommandType, MappingConstruct):
    __tag__ = '!Command'
    __rename_to__ = ImmutableDict([
        ('network-name', 'network_name'),
    ])

    name = attr.ib()  # type: str
    image = attr.ib()  # type: Union[str, DockerImage]
    run = attr.ib()  # type: Union[str, Sequence[str]]
    params = attr.ib(default=None)  # type: Optional[Sequence[ParameterType]]
    volumes = attr.ib(default=None)  # type: Optional[Sequence[VolumeType]]
    ports = attr.ib(default=None)  # type: Optional[Sequence[Expose]]
    environ = attr.ib(default=None)  # type: Optional[Mapping[str: str]]
    requires = attr.ib(default=None)  # type: Optional[Sequence[str]]
    network_name = attr.ib(default=None)  # type: Optional[str]
    description = attr.ib(default=None)  # type: Optional[str]

    def accept(self, visitor):
        return visitor.visit_command(self)


class ActionType:

    def accept(self, visitor):
        raise NotImplementedError


@attr.s
class Download(ActionType, ScalarConstruct):
    __tag__ = '!Download'

    url = attr.ib()

    def accept(self, visitor):
        return visitor.visit_download(self)


@attr.s
class File(ActionType, ScalarConstruct):
    __tag__ = '!File'

    path = attr.ib()

    def accept(self, visitor):
        return visitor.visit_file(self)


@attr.s
class Bundle(ActionType, ScalarConstruct):
    __tag__ = '!Bundle'

    path = attr.ib()

    def accept(self, visitor):
        return visitor.visit_bundle(self)
